import numpy as np
import scipy as sp
from comfit.tool.tool_print_in_color import tool_print_in_color
from comfit.tool.tool_configure_axis import tool_configure_axis

class BaseSystemInit:
    """ Initialization methods for the base system class"""
    def __init__(self, dim: int, **kwargs):
        """
        Initialize the class with the given parameters.

        Args:
            dimension (int): The dimension of the system. Must be 1, 2, or 3.
            **kwargs: Additional keyword arguments, see 
            https://comfitlib.com/ClassBaseSystem/

        Returns:
            None
        """

        if dim not in [1, 2, 3]:
            raise ValueError('Dimension must be 1, 2, or 3.')

        self.dim = dim

        # Checking what x-values were provided
        self.xlim = kwargs.get('xlim', None)
        self.xmin = kwargs.get('xmin', None)
        self.xmax = kwargs.get('xmax', None)
        self.xRes = kwargs.get('xRes', None)
        self.dx = kwargs.get('dx', None)

        self.xlim, self.xmin, self.xmax, self.xRes, self.dx = tool_configure_axis(self.dim, 'x',self.xlim, self.xmin, self.xmax, self.xRes, self.dx)

        # Checking what y-values were provided
        self.ylim = kwargs.get('ylim', None)
        self.ymin = kwargs.get('ymin', None)
        self.ymax = kwargs.get('ymax', None)
        self.yRes = kwargs.get('yRes', None)
        self.dy = kwargs.get('dy', None)

        self.ylim, self.ymin, self.ymax, self.yRes, self.dy = tool_configure_axis(self.dim, 'y',self.ylim, self.ymin, self.ymax, self.yRes, self.dy)

        # Checking what z-values were provided
        self.zlim = kwargs.get('zlim', None)
        self.zmin = kwargs.get('zmin', None)
        self.zmax = kwargs.get('zmax', None)
        self.zRes = kwargs.get('zRes', None)
        self.dz = kwargs.get('dz', None)

        self.zlim, self.zmin, self.zmax, self.zRes, self.dz = tool_configure_axis(self.dim, 'z',self.zlim, self.zmin, self.zmax, self.zRes, self.dz)

        # Minimum and maximum sizes of the domain
        self.size_x = self.xmax - self.xmin
        self.size_y = self.ymax - self.ymin if self.dim > 1 else 1
        self.size_z = self.zmax - self.zmin if self.dim > 2 else 1

        self.volume = self.size_x * self.size_y * self.size_z

        if self.dim == 1:
            self.size_min = self.size_x
            self.size_max = self.size_x
        elif self.dim == 2:
            self.size_min = np.min([self.size_x, self.size_y])
            self.size_max = np.max([self.size_x, self.size_y])
        elif self.dim == 3:
            self.size_min = np.min([self.size_x, self.size_y, self.size_z])
            self.size_max = np.max([self.size_x, self.size_y, self.size_z])
        

        #  Setting default value of time
        self.time = kwargs.get('time', 0)
        self.dt = kwargs.get('dt', 0.1)

        # Construct parameters
        self.x = np.linspace(self.xmin, self.xmax-self.dx, self.xRes)
        self.y = np.linspace(self.ymin, self.ymax-self.dy, self.yRes)
        self.z = np.linspace(self.zmin, self.zmax-self.dz, self.zRes)

        self.Res = self.xRes * self.yRes * self.zRes
        if self.dim == 1:
            self.dims = self.xRes
        elif self.dim == 2:
            self.dims = [self.xRes, self.yRes]
        elif self.dim == 3:
            self.dims = [self.xRes, self.yRes, self.zRes]

        self.a0 = 1  # System length scale

        # Helpful midpoints and their indices
        self.xmidi = (self.xRes) // 2 
        self.xmid = self.x[self.xmidi]

        self.ymidi = (self.yRes) // 2 
        self.ymid = self.y[self.ymidi]

        self.zmidi = (self.zRes) // 2 
        self.zmid = self.z[self.zmidi]

        self.midi = self.xRes * self.yRes * (self.zmidi - 1) + self.yRes * (self.xmidi - 1) + self.ymidi
        if self.dim == 1:
            self.rmid = self.xmid
            self.zero_index = 0 
        elif self.dim == 2:
            self.rmid = [self.xmid, self.ymid]
            self.zero_index = (0,0)
        elif self.dim == 3:
            self.rmid = [self.xmid, self.ymid, self.zmid]
            self.zero_index = (0,0,0)

        # Fourier modes
        self.k = [self.calc_wavenums(self.x)]
        if self.dim == 2:
            self.k[0] = self.k[0].reshape(self.xRes, 1)
            self.k.append(self.calc_wavenums(self.y).reshape(1, self.yRes))
        elif self.dim == 3:
            self.k[0] = self.k[0].reshape(self.xRes, 1, 1)
            self.k.append(self.calc_wavenums(self.y).reshape(1, self.yRes, 1))
            self.k.append(self.calc_wavenums(self.z).reshape(1, 1, self.zRes))

        # Reshape the position vectors
        if self.dim == 2:
            self.x = self.x.reshape((self.xRes, 1))
            self.y = self.y.reshape((1, self.yRes))
        elif self.dim == 3:
            self.x = self.x.reshape((self.xRes, 1, 1))
            self.y = self.y.reshape((1, self.yRes, 1))
            self.z = self.z.reshape((1, 1, self.zRes))


        # Derivatives
        self.dif = [1j * ki for ki in self.k]

        self.dV = self.dx
        if self.dim > 1:
            self.dV *= self.dy
        if self.dim > 2:
            self.dV *= self.dz

        self.rmin = [self.xmin, self.ymin, self.zmin]
        self.rmax = [self.xmax, self.ymax, self.zmax]


        # Plot lib
        self.plot_lib = kwargs.get('plot_lib', 'plotly')

    def __str__(self) -> str:
        """Return a string representation of the class.
        
        Input:
            None

        Returns:
            A string representation of the class.
        """
        description = f"BaseSystem instance\n"
        # Start with the dimension of the system
        description = f"System Dimension: {self.dim}\n"

        # Add x-axis properties
        description += f"X-Axis Limits: [{self.xmin}, {self.xmax}], Resolution: {self.xRes}, Delta: {self.dx}\n"

        # Add y-axis properties if dim > 1
        if self.dim > 1:
            description += f"Y-Axis Limits: [{self.ymin}, {self.ymax}], Resolution: {self.yRes}, Delta: {self.dy}\n"

        # Add z-axis properties if dim > 2
        if self.dim > 2:
            description += f"Z-Axis Limits: [{self.zmin}, {self.zmax}], Resolution: {self.zRes}, Delta: {self.dz}\n"

        # Add time properties
        description += f"Current Time: {self.time}, Time Step: {self.dt}"

        return description




from .tool_colormaps import tool_colormap_bluewhitered, tool_colormap_angle, tool_colormap_sunburst
from .tool_export_plot_functions import tool_export_rotating_plot
from .tool_plot_manipulation_functions import tool_zoom_plot
from .tool_math_functions import tool_multinom

from .tool_set_plot_axis_properties_matplotlib import tool_set_plot_axis_properties_matplotlib
from .tool_set_plot_axis_properties_plotly import tool_set_plot_axis_properties_plotly

from .tool_complete_field import tool_complete_field

from .tool_add_spacing_2D import tool_add_spacing_2D
from .tool_add_spacing_3D import tool_add_spacing_3D

from .tool_print_in_color import tool_print_in_color
from .tool_configure_axis import tool_configure_axis

from .tool_animation import tool_make_animation_movie, tool_make_animation_gif



from comfit.core.base_system_init import BaseSystemInit
from comfit.core.base_system_conf import BaseSystemConf
from comfit.core.base_system_evolve import BaseSystemEvolve
from comfit.core.base_system_calc import BaseSystemCalc
from comfit.core.base_system_plot import BaseSystemPlot
from comfit.core.base_system_get import BaseSystemGet

import scipy as sp

class BaseSystem(BaseSystemInit, BaseSystemConf, BaseSystemEvolve, BaseSystemCalc, BaseSystemPlot, BaseSystemGet):
    """
    The BaseSystem class is the base class for all systems in ComFiT.
    """
    
    def fft(self, field):
        """
        Perform a fast Fourier transform on a field. 

        Args: 
            field (array): The field to transform.
        Returns:
            array: The transformed field
        """
        
        return sp.fft.fftn(field, axes=range(-self.dim, 0))

    def ifft(self, field):
        """
        Perform an inverse fast Fourier transform on a field.

        Args:
            field (array): The field to transform
        
        Return:
            array: The transformed field
        """
        
        return sp.fft.ifftn(field, axes=range(-self.dim, 0))


import numpy
import numpy as np
from comfit.core.base_system import BaseSystem
import scipy as sp
from tqdm import tqdm
import matplotlib.pyplot as plt

from comfit.plot.plot_nodes_plotly import plot_nodes_plotly

from comfit.tool.tool_colormaps import tool_colormap_angle, tool_colormap_bluewhitered, tool_colormap_sunburst
from comfit.tool.tool_create_orthonormal_triad import tool_create_orthonormal_triad

from comfit.nematic_liquid_crystal.plot_field_velocity_and_director_matplotlib import plot_field_velocity_and_director_matplotlib

from comfit.nematic_liquid_crystal.plot_field_velocity_and_director_plotly import plot_field_velocity_and_director_plotly



from comfit.tool.tool_math_functions import levi_civita_symbol
from mpl_toolkits.mplot3d import axes3d
import matplotlib.cm as cm
from matplotlib.colors import Normalize

class NematicLiquidCrystal(BaseSystem):

    def __init__(self, dim, **kwargs):
        """Initializes a system to simulate a (active) nematic liquid crystal

        Args:
            dimension : The dimension of the system.
            kwargs : dict, optional
            Optional keyword arguments to set additional parameters. See
            https://comfitlib.com/ClassNematicLiquidCrystal/

        Returns:
        - NematicLiquidCrystal object
            The system object representing the nematic simulation.

        Example:
            nematic = NematicLiquidCrystal(2, xRes=100, yRes = 100, alpha=-0.5)
            Creates a nematic liquid crystal with 2 dimensions and a spatial resolution of 100.
            The activity alpha is set to -0.5.
        """
        super().__init__(dim, **kwargs)

        # Type of the system
        self.Q = None
        self.Q_f = None
        self.u = None
        self.u_f = None
        self.type = 'NematicLiquidCrystal'

        #defoult parameters
        self.alpha = -1 if 'alpha' not in kwargs else kwargs['alpha']   #defult is an extensile system
        self.K = 1 if 'K' not in kwargs else kwargs['K']
        self.A = 1 if 'A' not in kwargs else kwargs['A']
        self.B = 1 if 'B' not in kwargs else kwargs['B']
        self.C = 0 if 'C' not in kwargs  else kwargs['C']  # note: in two dim C is never called
        self.Lambda = 0 if 'Lambda' not in kwargs else kwargs['Lambda'] #flow allignment, not sure if this will be implemented
        self.gamma = 1  if 'gamma' not in kwargs else kwargs['gamma']  # rotational diffusion
        self.Gamma = 0 if 'Gamma' not in kwargs else kwargs['Gamma'] # friction,
        self.eta = 1 if 'eta' not in kwargs else kwargs['eta'] # viscosity


        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        """Output a string representation of the system.

        Input:
            None

        Returns:
            A string representation of the system.
        """
        return "NematicLiquidCrystal"

    # Initial condition
    def conf_initial_condition_ordered(self, noise_strength=0.01):
        """Configures the system with the nematogens pointing in the x-direction in 2D and in the z-direction in 3D
        with some random noise in the angle.
        
        Args:
             noise_strength: A meshure for how much noise to put in the angle (float)
        
        Returns:
            Initialises self.Q and self.Q_f
        
        Raises:
            Exception if the dimension is not 2 or 3
        """
        if self.dim == 2:
            S0 = np.sqrt(self.B)
            theta_rand = noise_strength*np.random.randn(self.xRes,self.yRes)
            self.Q = np.zeros((2,self.xRes,self.yRes))
            self.Q[0] = S0/2 *np.cos(2*theta_rand)
            self.Q[1] =  S0/2 *np.sin(2*theta_rand)


            self.Q_f = sp.fft.fft2(self.Q)

            self.k2 = self.calc_k2() # k2
            self.k2_press = self.calc_k2()
            self.k2_press[0,0] = 1 # for calculating pressure and velocity

        elif self.dim == 3:

            S0 = 1/8* self.C/self.A + 1/2 * np.sqrt(self.C**2 /(16*self.A**2) + 3*self.B)


            theta_rand = noise_strength*np.random.randn(self.xRes,self.yRes,self.zRes)
            phi_rand = noise_strength*np.random.randn(self.xRes,self.yRes,self.zRes)

            nx = np.cos(theta_rand)*np.sin(phi_rand)
            ny = np.sin(theta_rand)*np.sin(phi_rand)
            nz = np.cos(phi_rand)

            self.Q = np.zeros((5, self.xRes, self.yRes, self.zRes))
            self.Q[0] = S0 *(nx*nx -1/3)
            self.Q[1] = S0 *(nx*ny)
            self.Q[2] = S0 *(nx*nz)
            self.Q[3] = S0 *(ny*ny -1/3)
            self.Q[4] = S0 *(ny*nz)

            self.Q_f = sp.fft.fftn(self.Q, axes=(range(-self.dim, 0)))

            self.k2 = self.calc_k2()  # k2
            self.k2_press = self.calc_k2()
            self.k2_press[0, 0, 0] = 1  # for calculating pressure and velocity

        else:
            raise Exception("This dimension is not included for the moment")

    def conf_insert_disclination_dipole(self, dipole_vector=None, dipole_position=None):
        """Sets the initial condition for a disclination dipole configuration in a 2-dimensional system.

        Args:
            None

        Returns:
            Configures self.Q and self.Q_f with a disclination dipole configuration.

        Raises:
            Exception: If the dimension of the system is not 2.
        """
        if not (self.dim == 2):
            raise Exception("The dimension of the system must be 2 for a disclination dipole configuration.")

        if self.Q is None:
            self.conf_initial_condition_ordered(noise_strength=0)

        if dipole_vector is None:
            dipole_vector = [(self.xmax -self.xmin)/ 3, 0]

        if dipole_position is None:
            dipole_position = self.rmid

        psi = (self.Q[0] + 1j*self.Q[1]) * np.exp(1j * self.calc_angle_field_vortex_dipole(dipole_vector, dipole_position))
        self.Q[0] = np.real(psi)
        self.Q[1] = np.imag(psi)
        self.Q_f = sp.fft.fft2(self.Q)

    def conf_initial_disclination_lines(self, position1=None,position2 = None):
        """Sets the initial condition for a disclination line in a 3-dimensional system.

        The dislocation is parralell to the z-axis

        Args:
            position1 (list): the position of the first dislocation. Only the position in the xy plane is used

        Returns:
            Sets the value of self.Q and self.Q_f
        """
        if not (self.dim == 3):
            raise Exception("The dimension of the system must be 3 for a disclination line configuration.")

        if position1 is None:
            position1 = [self.xmid+(self.xmax-self.xmin)/3, self.ymid]
        if position2 is None:
            position2 = [self.xmid - (self.xmax-self.xmin) / 3, self.ymid]


        theta_1 = 1/2*np.arctan2((self.y-position1[1]),(self.x-position1[0]))
        theta_2 = -1/2*np.arctan2((self.y-position2[1]),(self.x-position2[0]))
        theta = theta_1 +theta_2

        S0 = 1/8* self.C/self.A + 1/2 * np.sqrt(self.C**2 /(16*self.A**2) + 3*self.B)

        nx =  np.cos(theta)
        ny = np.sin(theta)
        nz = np.zeros_like(nx)

        self.Q = np.zeros((5, self.xRes, self.yRes, self.zRes))
        self.Q[0] = S0 * (nx * nx - 1 / 3)
        self.Q[1] = S0 * (nx * ny)
        self.Q[2] = S0 * (nx * nz)
        self.Q[3] = S0 * (ny * ny - 1 / 3)
        self.Q[4] = S0 * (ny * nz)

        self.Q_f = sp.fft.fftn(self.Q, axes=(range(-self.dim, 0)))

        self.k2 = self.calc_k2()  # k2
        self.k2_press = self.calc_k2()
        self.k2_press[0, 0, 0] = 1




    def conf_active_channel(self,width = None,d=7):
        """Configures the activity to zero everywhere exept for inside a channel of width "width"
        
        Args:
            width: width of the channel (float)
            d: width of interface (float, optional)
        
        Returns:
            Updates the activity to the channel configuration.
        """
        if self.dim ==2:
            X, Y = np.meshgrid(self.x, self.y, indexing='ij')
            alpha_0 = self.alpha
            if width is None:
                width = (self.xmax-self.xmin)/3

            self.alpha = alpha_0*(1- 1 / 2 * (2 + np.tanh((X - self.xmid - width/2) / d) - np.tanh((X - self.xmid + width/2) / d)))
        else:
            raise Exception("The active channel is only permitted in two dimensions")

    ## calculations related to the flow field
    def conf_velocity(self,Q):
        '''
        Updates the velocity and its fourier transform given a nematic field Q.
        
        Args:
            (numpy.narray) the Q tensor
        
        Returns:
            (numpy.narray) velocity
        '''
        F_af = self.calc_active_force_f(Q)
        F_pf = self.calc_passive_force_f(Q)
        p_f = self.calc_pressure_f(F_af,F_pf)
        grad_pf = self.calc_gradient_pressure_f(p_f)
        self.u_f = (F_af + F_pf-grad_pf )/ (self.Gamma +self.eta*self.k2_press)

        self.u = np.real(sp.fft.ifftn(self.u_f, axes=(range(-self.dim, 0))))

    def calc_active_force_f(self,Q):
        '''Function that calculates the activ force in Fourier space.
        
        Args:
            Q: the order parameter that we use to find the force.  (numpy.narray) 
        
        Returns:
            The active force in Fourier space (numpy.narray) 
        '''
        F_af = []
        for j in range(self.dim):
            F_af.append(np.sum(1j*self.k[i]*sp.fft.fftn(self.alpha *self.get_sym_tl(Q,j,i),axes=(range(-self.dim, 0)) ) for i in range(self.dim)))
        return np.array(F_af)

    def calc_passive_force_f(self,Q):
        '''Calculates the passive force in Fourier space
        
        Args:
            Q: the order parameter that we use to find the force. (numpy.narray)
        
        Returns: 
            The passive force in Fourier space numpy.ndarray: 
        '''
        Pi_f = self.calc_passive_stress_f(Q)
        F_pf = []
        for j in range(self.dim):
            F_pf.append(np.sum(1j * self.k[i] *Pi_f[j][i] for i in range(self.dim)))
        return numpy.array(F_pf)

    def calc_passive_stress_f(self,Q):
        """Calculates the passive stress in fourier space

        Args:
            Q: the order parameter that we use to find the stress.
        
        Returns: 
            The passive stress in fourier space (numpy.narray) 
        """
        if self.dim == 2:
            H = self.calc_molecular_field(Q)
            Antisym_QH = np.sum(self.get_sym_tl(Q,0,k)*self.get_sym_tl(H,k,1) -self.get_sym_tl(H,0,k)*self.get_sym_tl(Q,k,1) for k in range(self.dim))
            Ericksen = np.zeros((self.dim,self.xRes,self.yRes),dtype=np.complex128)
            Ericksen[0] = - self.K*np.sum(sp.fft.ifftn(1j*self.k[0]*sp.fft.fftn(self.get_sym_tl(Q,m,l)))*
                                              sp.fft.ifftn(1j * self.k[0] * sp.fft.fftn(self.get_sym_tl(Q,m,l)))
                                              for m in range(self.dim) for l in range(self.dim))
            Ericksen[1] = - self.K*np.sum(sp.fft.ifftn(1j*self.k[0]*sp.fft.fftn(self.get_sym_tl(Q,m,l)))*
                                              sp.fft.ifftn(1j * self.k[1] * sp.fft.fftn(self.get_sym_tl(Q,m,l)))
                                              for m in range(self.dim) for l in range(self.dim))

            stress = np.zeros((self.dim,self.dim,self.xRes,self.yRes),dtype=np.complex128)
            for i in range(self.dim):
                for j in range(self.dim):
                    stress[i][j] = self.get_sym_tl(Ericksen,i,j) + self.get_anti_sym(Antisym_QH,i,j)
            return sp.fft.fftn(stress, axes=(range(-self.dim, 0)) )

        elif self.dim == 3:
            # TODO optimise
            H = self.calc_molecular_field(Q)

            Antisym_QH = np.zeros((3, self.xRes, self.yRes,self.zRes), dtype=np.complex128)

            Antisym_QH[0] = np.sum(self.get_sym_tl(Q,0,k)*self.get_sym_tl(H,k,1) -self.get_sym_tl(H,0,k)*self.get_sym_tl(Q,k,1) for k in range(self.dim))
            Antisym_QH[1] = np.sum(
                self.get_sym_tl(Q, 0, k) * self.get_sym_tl(H, k, 2) - self.get_sym_tl(H, 0, k) * self.get_sym_tl(Q, k, 2) for k in
                range(self.dim))
            Antisym_QH[2] = np.sum(
                self.get_sym_tl(Q, 1, k) * self.get_sym_tl(H, k, 2) - self.get_sym_tl(H, 1, k) * self.get_sym_tl(Q, k, 2) for k in
                range(self.dim))

            Ericksen = np.zeros((5, self.xRes, self.yRes,self.zRes), dtype=np.complex128)
            Ericksen[0] = - self.K * np.sum(sp.fft.ifftn(1j * self.k[0] * sp.fft.fftn(self.get_sym_tl(Q, m, l))) *
                                            sp.fft.ifftn(1j * self.k[0] * sp.fft.fftn(self.get_sym_tl(Q, m, l)))
                                            for m in range(self.dim) for l in range(self.dim))
            Ericksen[1] = - self.K * np.sum(sp.fft.ifftn(1j * self.k[0] * sp.fft.fftn(self.get_sym_tl(Q, m, l))) *
                                            sp.fft.ifftn(1j * self.k[1] * sp.fft.fftn(self.get_sym_tl(Q, m, l)))
                                            for m in range(self.dim) for l in range(self.dim))
            Ericksen[2] = - self.K * np.sum(sp.fft.ifftn(1j * self.k[0] * sp.fft.fftn(self.get_sym_tl(Q, m, l))) *
                                            sp.fft.ifftn(1j * self.k[2] * sp.fft.fftn(self.get_sym_tl(Q, m, l)))
                                            for m in range(self.dim) for l in range(self.dim))
            Ericksen[3] = - self.K * np.sum(sp.fft.ifftn(1j * self.k[1] * sp.fft.fftn(self.get_sym_tl(Q, m, l))) *
                                            sp.fft.ifftn(1j * self.k[1] * sp.fft.fftn(self.get_sym_tl(Q, m, l)))
                                            for m in range(self.dim) for l in range(self.dim))
            Ericksen[4] = - self.K * np.sum(sp.fft.ifftn(1j * self.k[1] * sp.fft.fftn(self.get_sym_tl(Q, m, l))) *
                                            sp.fft.ifftn(1j * self.k[2] * sp.fft.fftn(self.get_sym_tl(Q, m, l)))
                                            for m in range(self.dim) for l in range(self.dim))

            stress = np.zeros((self.dim,self.dim,self.xRes,self.yRes,self.zRes))
            for i in range(self.dim):
                for j in range(self.dim):
                    stress[i][j] = self.get_sym_tl(Ericksen,i,j) + self.get_anti_sym(Antisym_QH,i,j)
            return sp.fft.fftn(stress, axes=(range(-self.dim, 0)) )

    def calc_trace_Q2(self,Q):

        if self.dim == 2:
            return 2*( Q[0]**2 + Q[1]**2)
        elif self.dim == 3:
            return 2 *(Q[0]**2 + Q[1]**2+ Q[2]**2+Q[3]**2 + Q[4]**2 + Q[0]*Q[3])


    def calc_molecular_field(self,Q):
        """Finds the molecular field 
        
        Args:
            Q (numpy.ndarray): The nematic tensor
        
        Returns:
            The molecular field (numpy.ndarray)
        """

        Q2 =  self.calc_trace_Q2(Q)
        temp = -self.K * sp.fft.ifftn( self.k2* sp.fft.fftn(Q,axes=(range(-self.dim,0))),axes=(range(-self.dim,0)) )

        if self.dim == 2 or self.C == 0:
            return temp +self.A*self.B*Q -2*self.A*Q2*Q

        elif self.dim == 3:
            C_term = np.zeros((5,self.xRes,self.yRes,self.zRes))

            C_term[0] = self.C * (Q[0]**2 + Q[1]**2 +Q[2]**2 - 1/3 *Q2)
            C_term[1] = self.C * (Q[0]*Q[1] + Q[1]*Q[3] +Q[2]*Q[4])
            C_term[2] = self.C * (Q[1]*Q[4] -Q[2] * Q[3] )
            C_term[3] = self.C * (Q[1]**2 + Q[3]**2 +Q[4]**2 - 1/3 *Q2)
            C_term[4] = self.C * (Q[1]*Q[2]  -Q[4]*Q[0] )

            return temp +self.A*self.B*Q -2*self.A*Q2*Q +C_term


    def calc_pressure_f(self,F_af,F_pf):
        '''Calculates the pressure in Fourier space. The zero mode is set to zero
        
        Args:
            F_af: the active force in Fourier space  (numpy.narray)
            F_pf:the passive force in Fourier space  (numpy.narray) 
        
        Returns: 
            The pressure (numpy.ndarray) 
        '''
        p_af = np.sum(1j*self.k[i]*F_af[i] for i in range(self.dim))
        p_pf = np.sum(1j*self.k[i]*F_pf[i] for i in range(self.dim))
        return -(p_af + p_pf)/self.k2_press

    
    def calc_gradient_pressure_f(self,p_f):
        """Caclulates the gradient of the pressure
        
        Args: 
            p_f: the pressure in Fourier space  (numpy.narray)
        
        Returns:
            Gradient of the pressure (numpy.ndarray) 
        """
        grad_pf = []
        for i in range(self.dim):
            grad_pf.append(1j*self.k[i]*p_f)
        return np.array(grad_pf)

    def calc_vorticity_tensor(self):
        """Calculates the vorticity tensor

        Args:
            None

        Returns:
            The vorticity tensor (numpy.ndarray) 
        """
        if self.dim == 2:

            Omega_f = (1j*self.k[0]*self.u_f[1] -1j*self.k[1]*self.u_f[0])/2
            Omega = sp.fft.ifftn(Omega_f,axes=range(-self.dim,0))
            return np.real(Omega)

        elif self.dim ==3:
            Omega = np.zeros((3,self.xRes,self.yRes,self.zRes))
            Omega[0] = np.real(sp.fft.ifftn((1j*self.k[0]*self.u_f[1] -1j*self.k[1]*self.u_f[0])/2,axes=range(-self.dim,0)))
            Omega[1] = np.real(sp.fft.ifftn((1j * self.k[0] * self.u_f[2] - 1j * self.k[2] * self.u_f[0]) / 2,
                                            axes=range(-self.dim, 0)))
            Omega[2] = np.real(sp.fft.ifftn((1j * self.k[1] * self.u_f[2] - 1j * self.k[2] * self.u_f[1]) / 2,
                                            axes=range(-self.dim, 0)))
            return Omega

    def calc_strain_rate_tensor_f(self):
        """
        Calculates the strainrate tensor

        Args:
            None

        Returns: 
            The strainrate (numpy.ndarray) 
        """
        trace_u = np.sum(1j*self.k[i]*self.u_f[i] for i in range(self.dim))
        if self.dim == 2:
            E_f = np.zeros((2,self.xRes,self.yRes),dtype=np.complex128)
            E_f[0] = (1j*self.k[0] *self.u_f[0]) - trace_u/2
            E_f[1] = (1j*self.k[0] *self.u_f[1] +1j*self.k[1] *self.u_f[0])/2
        elif self.dim == 3:
            E_f = np.zeros((5, self.xRes, self.yRes,self.zRes), dtype=np.complex128)
            E_f[0] = (1j*self.k[0] *self.u_f[0]) - trace_u/3
            E_f[1] = (1j*self.k[0] *self.u_f[1] +1j*self.k[1] *self.u_f[0])/2
            E_f[2] = (1j * self.k[0] * self.u_f[2] + 1j * self.k[2] * self.u_f[0]) / 2
            E_f[3] =  (1j*self.k[1] *self.u_f[1]) - trace_u/3
            E_f[4] = (1j * self.k[1] * self.u_f[2] + 1j * self.k[2] * self.u_f[1]) / 2
        return E_f

    ## Calculation of non-linear evolution terms
    def calc_nonlinear_evolution_function_f(self,Q,t):
        """Calculates the non-linear evolution function for the nematic
        
        Args:
            Q: the nematic orderparameter (numpy.narray) 
        
        Returns:
            
            The non-linear evolution function evaluated in Fourier space (numpy.narray) 
        """
        if self.dim == 2:
            self.conf_velocity(Q)
            Q_f = sp.fft.fftn(Q,axes=range(-self.dim,0))
            N_f = self.calc_nonlinear_evolution_term_no_flow_f(Q,t)
            Omega =self.calc_vorticity_tensor()
            Antisym_Omega_Q = np.zeros_like(Q_f)

            Antisym_Omega_Q[0] = -2 *Q[1]*Omega
            Antisym_Omega_Q[1] = 2*Q[0]*Omega

            advectiv_deriv = - np.sum(self.u[k]* sp.fft.ifftn(1j*self.k[k] * Q_f,axes=(range(-self.dim,0)))for k in range(self.dim) )

            return sp.fft.fftn(Antisym_Omega_Q +advectiv_deriv, axes=range(-self.dim,0)) +N_f

        elif self.dim == 3:
            self.conf_velocity(Q)
            Q_f = sp.fft.fftn(Q, axes=range(-self.dim, 0))
            N_f = self.calc_nonlinear_evolution_term_no_flow_f(Q,t)
            Omega = self.calc_vorticity_tensor()

            advectiv_deriv = - np.sum(
                self.u[k] * sp.fft.ifftn(1j * self.k[k] * Q_f, axes=(range(-self.dim, 0))) for k in range(self.dim))

            Antisym_Omega_Q = np.zeros_like(Q_f)

            Antisym_Omega_Q[0] = -2 * (Q[1]*Omega[0] + Q[2]*Omega[1] )
            Antisym_Omega_Q[1] = Omega[0]*(Q[0]-Q[3]) - Omega[2]*Q[2] -Omega[1]*Q[4]
            Antisym_Omega_Q[2] = Omega[1]*(2*Q[0] + Q[3] ) + Omega[1]*Q[1] - Omega[0]*Q[4]
            Antisym_Omega_Q[3] = 2*(Q[1]*Omega[0] -Q[4]*Omega[2])
            Antisym_Omega_Q[4] = Omega[2]*(2*Q[3]+Q[0]) + Omega[1]*Q[1] + Omega[0]*Q[2]

            return sp.fft.fftn(Antisym_Omega_Q + advectiv_deriv, axes=range(-self.dim, 0)) + N_f

        else:
            raise Exception("This dimension is not implemented at the moment")

    def calc_nonlinear_evolution_term_no_flow_f(self,Q,t):
        """Calculates the non-linear evolution function for the nematic without the flow field
        
        Args:
            Q: the nematic orderparameter  (numpy.narray)
        
        Returns:
            The non-linear evolution function evaluated in Fourier space (numpy.narray) 
        """
        Q2 = self.calc_trace_Q2(Q)
        if self.dim ==2 or self.C == 0:
            return -2*self.A*sp.fft.fftn(Q2 *Q,axes =(range(-self.dim,0)))/self.gamma

        elif self.dim == 3:
            C_term = np.zeros((5, self.xRes, self.yRes, self.zRes))

            C_term[0] = self.C * (Q[0] ** 2 + Q[1] ** 2 + Q[2] ** 2 - 1 / 3 * Q2)
            C_term[1] = self.C * (Q[0] * Q[1] + Q[1] * Q[3] + Q[2] * Q[4])
            C_term[2] = self.C * (Q[1] * Q[4] - Q[2] * Q[3])
            C_term[3] = self.C * (Q[1] ** 2 + Q[3] ** 2 + Q[4] ** 2 - 1 / 3 * Q2)
            C_term[4] = self.C * (Q[1] * Q[2] - Q[4] * Q[0])
            return -2*self.A*sp.fft.fftn(Q2 *Q,axes =(range(-self.dim,0)))/self.gamma + sp.fft.fftn(C_term ,axes =(range(-self.dim,0)))/self.gamma


##### evolvers
    def evolve_nematic(self, number_of_steps, method= 'ETD2RK'):
        '''Evolves the nematic system
        
        Args:
            number_of_steps: the number of time steps that we are evolving the equation  (int)
            method: the integration method we want to use. ETD2RK is sett as default  (string)
        
        Returns:
            Updates the fields self.Q and self.Q_f
        '''
        omega_f = (self.A * self.B - self.K * self.k2) / self.gamma

        if method == 'ETD2RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD2RK(omega_f)
            solver = self.evolve_ETD2RK_loop

        elif method == 'ETD4RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD4RK(omega_f)
            solver = self.evolve_ETD4RK_loop
        else:
            raise Exception('Mehtod is not implemented')
        for n in tqdm(range(number_of_steps), desc='evolving the active nematic'):
            self.Q, self.Q_f = solver(integrating_factors_f,
                                                       self.calc_nonlinear_evolution_function_f,
                                                       self.Q, self.Q_f)
        self.Q = np.real(self.Q)
        self.conf_velocity(self.Q)

    def evolve_nematic_no_flow(self,number_of_steps,method = 'ETD2RK'):
        ''' Evolves the nematic system without the flow field
        
        Args:
            number_of_steps: the number of time steps that we are evolving the equation  (int)
            method: the integration method we want to use. ETD2RK is sett as default  (string)
        
        Returns:
            Updates the fields self.Q and self.Q_f
        '''
        omega_f = (self.A * self.B - self.K * self.k2) / self.gamma

        if method == 'ETD2RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD2RK(omega_f)
            solver = self.evolve_ETD2RK_loop

        elif method == 'ETD4RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD4RK(omega_f)
            solver = self.evolve_ETD4RK_loop
        else:
            raise Exception('Mehtod is not implemented')

        for n in tqdm(range(number_of_steps), desc='evolving the active nematic'):
            self.Q, self.Q_f = solver(integrating_factors_f,self.calc_nonlinear_evolution_term_no_flow_f,
                                                           self.Q, self.Q_f)
        self.Q = np.real(self.Q)


    ## Disclination tracking
    def calc_disclination_density_nematic(self):
        """
        Calculates the disclination density for the nematic. Note that in three dimension the disclination density is a tensor

        Args:
            None

        Returns:
            The disclination density (numpy.narray) 
        """
        if self.dim == 2:
            psi0 = np.sqrt(self.B)/2
            psi =[self.Q[0],self.Q[1]]
            return self.calc_defect_density(psi, psi0)

        elif self.dim == 3:
            D = np.zeros((self.dim,self.dim,self.xRes,self.yRes,self.zRes))
            term_trace = np.sum(np.real(sp.fft.ifftn(1j*self.k[k]* self.get_sym_tl(self.Q_f,k,a)))*
                                    np.real(sp.fft.ifftn(1j*self.k[l] * self.get_sym_tl(self.Q_f,l,a)))
                                    - np.real(sp.fft.ifftn(1j*self.k[k]* self.get_sym_tl(self.Q_f,l,a)))*
                                    np.real(sp.fft.ifftn(1j*self.k[l] * self.get_sym_tl(self.Q_f,k,a)))
                                    for k in range(self.dim) for a in range(self.dim) for l in range(self.dim))

            for gam in range(self.dim):
                for i in range(self.dim):
                    D[gam, i] = 2*np.sum(np.real(sp.fft.ifftn(1j*self.k[gam]* self.get_sym_tl(self.Q_f,k,a)))*
                                    np.real(sp.fft.ifftn(1j*self.k[k] * self.get_sym_tl(self.Q_f,i,a)))
                                    - np.real(sp.fft.ifftn(1j*self.k[gam]* self.get_sym_tl(self.Q_f,i,a)))*
                                    np.real(sp.fft.ifftn(1j*self.k[k] * self.get_sym_tl(self.Q_f,k,a)))
                                    for k in range(self.dim) for a in range(self.dim))
                    if gam == i:
                        D[gam,i] += term_trace
            S0 = self.calc_equilibrium_S()
            return D/(S0**2 *np.pi)
        else:
            raise Exception("Only two and three dimensions are currently supported")


    def calc_disclination_density_decoupled(self):
        """Calculates the decoupled disclination density

        Args:
            None
        
        Returns:
            The disclination density (numpy.narray)
        """ 
        if self.dim == 3:
            rho = self.calc_disclination_density_nematic()


            omega = np.sqrt(np.sum(rho[i,j]*rho[i,j] for i in range(self.dim) for j in range(self.dim)) )

            DDT = np.zeros((self.xRes,self.yRes,self.zRes,self.dim,self.dim))
            DTD = np.zeros((self.xRes,self.yRes,self.zRes,self.dim,self.dim))

            for i in range(self.dim):
                for j in range(self.dim):
                    DDT[:,:,:,i,j] = np.sum(rho[i,k]*rho[j,k] for k in range(self.dim))
                    DTD[:, :, :, i, j] = np.sum(rho[k,i] * rho[ k,j] for k in range(self.dim))

            vals_1,vecs_1 =  numpy.linalg.eigh(DDT)
            vals_2, vecs_2 = numpy.linalg.eigh(DTD)

            Omega = np.transpose(vecs_1[:,:,:,:,2], (3,0,1,2))
            T = np.transpose(vecs_2[:,:,:,:,2], (3,0,1,2))
            trRho = np.sum(rho[i,i] for i in range(self.dim))
            return omega, Omega, T, trRho


    def calc_dt_psi(self,Q_prev,delta_t):
        """Calculates the time derivative of the order parameter as a complex field

        Args:
            Q_prev: the order parameter at the previous time step (numpy.narray)
            delta_t: the time step (float)
        
        Returns:
            The time derivative of the order parameter (numpy.narray)
        """

        dt_Q = (self.Q -Q_prev)/delta_t
        return dt_Q[0] + 1j*dt_Q[1]

    def calc_equilibrium_S(self):
        '''Calculates the strength of nematic order S

        Args:
            None

        Returns:
            equilibrium value of S (numpy.narray)
        '''
        if self.dim == 2:
            return  np.sqrt(self.B)

        elif self.dim == 3:
            S0 = 1 / 8 * self.C / self.A + 1 / 2 * np.sqrt(self.C ** 2 / (16 * self.A ** 2) + 3 * self.B)
            return S0

    def calc_order_and_director(self):
        """Calculates the amount of order (S) and the director field (n)

        Args:
            None

        Returns:
            Tuple consisting of 
                - Amount of order (scalar field) 
                - the director field (vector field)
        """
        if self.dim == 2:
            psi_n = self.Q[0] + 1j*self.Q[1]
            angle = np.angle(psi_n)
            S =2 * np.sqrt((self.Q[0]) ** 2 + (self.Q[1]) ** 2)
            return S, [np.cos(angle/2),np.sin(angle/2)]
        elif self.dim ==3:
        ## need to construct a matrix
            Q_eig = numpy.zeros((self.xRes,self.yRes,self.zRes,self.dim,self.dim))
            for i in range(self.dim):
                for j in range(self.dim):
                    Q_eig[:,:,:,i,j] = self.get_sym_tl(self.Q,i,j)

            eigvals, eigvectors = numpy.linalg.eigh(Q_eig)
            S = 3/2 *eigvals[:,:,:,2]
            n = np.transpose(eigvectors[:,:,:,:,2], (3,0,1,2))

            return S, n

    def calc_disclination_velocity_field(self, dt_Q):
        """
        Calculates the velocity field of the disclination in two dimensions
        
        Args:
            dt_Q: the time derivative of the order parameter (numpy.narray)

        Returns:
            The velocity field (numpy.narray) 
        """
        if self.dim ==2:

            psi = self.Q[0] +1j*self.Q[1]

            dt_psi = dt_Q[0] + 1j * dt_Q[1]

            return self.calc_defect_velocity_field([np.real(psi), np.imag(psi)],
                                                   [np.real(dt_psi), np.imag(dt_psi)])


    def calc_disclination_polarization_field(self):
        """Calculates the polarization field of the disclination in two dimensions

        Args:
            None
        
        Returns:
            The polarization field (numpy.narray)
        """
        ex = np.real(sp.fft.ifftn(1j*self.k[0]*self.Q_f[0] + 1j*self.k[1]*self.Q_f[1]))
        ey = np.real(sp.fft.ifftn(1j * self.k[0] * self.Q_f[1] - 1j * self.k[1] * self.Q_f[0]))
        return np.array([ex,ey])

    def calc_disclination_nodes_nem(self, dt_Q=None,polarization = None,charge_tolerance=None):
        """Calculates the positions and charges of disclination nodes based on the disclination density.
        
        Args:
            dt_Q: The time derivative of the order parameter. If not provided, the velocity of the disclination nodes will not be calculated. (numpy.narray, optional)
        
        Returns:
            A list of dictionaries representing the disclination nodes. Each dictionary contains the following keys:
                  - 'position_index': The position index of the disclination node in the disclination density array.
                  - 'charge': The charge of the disclination node.
                  - 'position': The position of the disclination node as a list [x, y].
                  - 'velocity': The velocity of the disclination node as a list [vx, vy].
        """

        # Calculate disclination density
        if self.dim == 2:

            rho = self.calc_disclination_density_nematic()

            if dt_Q is not None:
                velocity_field = self.calc_disclination_velocity_field(dt_Q)


            disclination_nodes = self.calc_defect_nodes(np.abs(rho))
            for disclination in disclination_nodes:
                disclination['charge'] = np.sign(rho[disclination['position_index']])

                if dt_Q is not None:
                    disclination['velocity'] = [velocity_field[0][disclination['position_index']],
                                          velocity_field[1][disclination['position_index']]]
                else:
                    disclination['velocity'] = [float('nan'), float('nan')]

                if disclination['charge'] >0 and polarization is not None:
                    disclination['polarization'] = [polarization[0][disclination['position_index']]
                        ,polarization[1][disclination['position_index']]]
                else:
                    disclination['polarization'] = [float('nan'), float('nan')]

        elif self.dim == 3:
            omega, Omega, T, trD = self.calc_disclination_density_decoupled()
            S0 = self.calc_equilibrium_S()
            disclination_nodes = self.calc_defect_nodes(omega,charge_tolerance=S0/np.pi)
            position_list = []

            for disclination in disclination_nodes:

                tangent_vector = np.array([T[i][disclination['position_index']] for i in range(3)])
                rotation_vector = np.array([Omega[i][disclination['position_index']] for i in range(3)])

                for i in range(len(position_list)):
                    pos = position_list[i]
                    if np.sqrt(sum( (disclination['position'][i] -pos[i])**2 for i in range(self.dim) )) <  5*self.a0:
                        tan_neight = disclination_nodes[i]['Tangent_vector']
                        if np.sum((tan_neight[j] -tangent_vector[j] )**2 -(tan_neight[j] + tangent_vector[j])**2 for j in range(self.dim)) > 0:

                            tangent_vector = -1* tangent_vector
                        break

                if np.sign(np.dot(tangent_vector, rotation_vector)) != np.sign(trD[disclination['position_index']]):
                    rotation_vector = -1*rotation_vector

                disclination['Tangent_vector'] = tangent_vector
                disclination['Rotation_vector'] = rotation_vector
                position_list.append(disclination['position'])


        return disclination_nodes

    def plot_field_velocity_and_director(self, field, velocity, director, **kwargs):
        if self.plot_lib == 'plotly':
            return plot_field_velocity_and_director_plotly(self, field, velocity, director, **kwargs)
        elif self.plot_lib == 'matplotlib':
            return plot_field_velocity_and_director_matplotlib(self, field, velocity, director, **kwargs)



from typing import Callable

import numpy as np
import scipy as sp

class BaseSystemEvolve:
    """ Evolution methods for the base system class"""
    ## Time evolution function
    def evolve_ETD2RK_loop(
        self, 
        integrating_factors_f: list[np.ndarray], 
        nonlinear_evolution_function_f: Callable[[np.ndarray, float], np.ndarray], 
        field: np.ndarray, 
        field_f: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Evolves the given field using the ETD2RK scheme with a loop.

        Args:
            integrating_factors_f: A list of three integrating factors. (list)
            nonlinear_evolution_function_f: A function that calculates the non-linear evolution of the field
                    and returns the fourier transform.  (function)
            field: The initial field to be evolved. (ndarray)
            field_f: The Fourier transform of the initial field. (ndarray)

        Returns:
            tuple: A tuple containing the evolved field and the predicted field in Fourier space.
        """

        N0_f = nonlinear_evolution_function_f(field, self.time)

        a_f = integrating_factors_f[0] * field_f + integrating_factors_f[1] * N0_f
        a = self.ifft(a_f)

        N_a_f = nonlinear_evolution_function_f(a, self.time+self.dt)
        field_f = a_f + integrating_factors_f[2] * (N_a_f - N0_f)
        field = self.ifft(field_f)

        self.time += self.dt

        return field, field_f

    def evolve_ETD4RK_loop(
        self, 
        integrating_factors_f: list[np.ndarray], 
        nonlinear_evolution_function_f: Callable[[np.ndarray, float], np.ndarray], 
        field: np.ndarray, 
        field_f: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Evolves the given field using the ETD4RK scheme with a loop.

         Args:
             integrating_factors_f: A list of five integrating factors.  (list)
             nonlinear_evolution_function_f: A function that calculates the non-linear evolution of the field.  (function)
             field: The initial field to be evolved. (ndarray)
             field_f: The Fourier transform of the initial field. (ndarray)

         Returns:
             A tuple containing the evolved field and the predicted field in Fourier space.
         """
         
        N_0f = nonlinear_evolution_function_f(field, self.time)

        a_f = field_f * integrating_factors_f[0] + N_0f * integrating_factors_f[1]
        a = self.ifft(a_f)
        N_a = nonlinear_evolution_function_f(a, self.time + self.dt / 2)

        b_f = field_f * integrating_factors_f[0] + N_a * integrating_factors_f[1]
        b = self.ifft(b_f)
        N_b = nonlinear_evolution_function_f(b, self.time + self.dt / 2)

        c_f = a_f * integrating_factors_f[0] + (2 * N_b - N_0f) * integrating_factors_f[1]
        c = self.ifft(c_f)
        N_c = nonlinear_evolution_function_f(c, self.time + self.dt)

        field_f = field_f * integrating_factors_f[2] + N_0f * integrating_factors_f[3] \
                  + (N_a + N_b) * integrating_factors_f[4] + N_c * integrating_factors_f[5]

        field = self.ifft(field_f)

        self.time += self.dt

        return field, field_f


import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from comfit.plot.plot_complex_field_matplotlib import plot_complex_field_matplotlib
from comfit.tool import tool_complete_field



def plot_angle_field_matplotlib(self, 
       angle_field: np.ndarray, 
        **kwargs
        ) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plot the angle field.

    Args:
        field (array-like): The angle field values.
        ax (matplotlib.axes.Axes, optional): The axes to plot the angle field on. If not provided, a new subplot will be created.
    
    Returns:
        tuple: A tuple containing:
            - matplotlib.figure.Figure: The figure containing the plot.
            - matplotlib.axes.Axes: The axes containing the plot.
    """

    # Check if the vector field is complex
    if np.iscomplexobj(angle_field):
        print("\033[91mWarning: the angle vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(angle_field)))
        angle_field = np.real(angle_field)

    # Extend the field if not a complete array is given
    angle_field = tool_complete_field(self, angle_field)

    # Normalize around 0
    angle_field = np.mod(angle_field + np.pi, 2 * np.pi) - np.pi        

    if self.dim == 1:
        if 'vlim' in kwargs:
            vlim = kwargs['vlim']
        else:
            kwargs['vlim'] = [-np.pi, np.pi]
            kwargs['yticks'] = [-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]
            kwargs['yticklabels'] = [r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$']

        
        return plot_field_matplotlib(self, angle_field, **kwargs)
    
    elif self.dim > 1:
        complex_field = np.exp(1j * angle_field)

        kwargs['plot_method'] = 'phase_angle'

        return plot_complex_field_matplotlib(self, complex_field, **kwargs)


import numpy as np
from comfit.core.base_system import BaseSystem
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint
import copy
from scipy.optimize import curve_fit

# Plot functions
from comfit.plot.plot_field_plotly import plot_field_plotly
from comfit.plot.plot_field_matplotlib import plot_field_matplotlib

from comfit.tool.tool_print_in_color import tool_print_in_color


class PhaseFieldCrystal(BaseSystem):

    def __init__(self, dim, **kwargs):
        """
        Initializes a system to simulate a Phase Field Crystal. This class is the base of the other
        phase field crystal models implemented in comfit

        Args:
            dim: The dimension of the system.
            kwargs :keyword arguments to set additional parameters. See https://comfitlib.com/ClassPhaseFieldCrystal/

        Returns:
            The system object representing the phase field crystal simulation.
        """
        
        # First initialize the BaseSystem class
        super().__init__(dim, **kwargs)

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)
            

        self.dislocation_charges = np.array(
            [[np.round(np.dot(an, qn) / (2 * np.pi), decimals=8) for qn in self.q] for an in self.a])

        # Elastic constant
        if ((self.dim - 1) * self.el_lambda + 2 * self.el_mu + self.el_gamma)!=0:
            self.el_nu = self.el_lambda / ((self.dim - 1) * self.el_lambda + 2 * self.el_mu + self.el_gamma)
        else:
            self.el_nu = 0

        self.Phi = 2*sum(np.array(self.eta0)**2)

    def __str__(self):
        """
        Returns a string representation of the object.

        Args:
            None
        
        Returns: 
            A string representation of the object.
        """
        string = "-------------------------------\n"
        string += "Phase Field Crystal instance\n"
        string += "-------------------------------\n"
        string += "Type: " + self.type + "\n"
        if self.dim == 1:
            string += "(nx) = (" + str(self.nx) + ")\n"
        elif self.dim == 2:
            string += "(nx, ny) = (" + str(self.nx) + ", " + str(self.ny) + ")\n"
        elif self.dim == 3:
            string += "(nx, ny, nz) = (" + str(self.nx) + ", " + str(self.ny) + ", " + str(self.nz) + ")\n"
        string += "micro_resolution: " + str(self.micro_resolution) + "\n"
        string += "-------------------------------\n"
        string += "Parameters\n"
        string += "-------------------------------\n"
        string += "r: " + str(self.r) + "\n"
        string += "t: " + str(self.t) + "\n"
        string += "v: " + str(self.v) + "\n"
        string += "-------------------------------\n"
        string += "Proto amplitudes\n"
        string += "-------------------------------\n"
        string += "psi0: " + str(self.psi0) + "\n"
        if self.type == 'PhaseFieldCrystal1DPeriodic':
            string += "A: " + str(self.A) + "\n"
        elif self.type == 'PhaseFieldCrystal2DTriangular':
            string += "A: " + str(self.A) + "\n"
        elif self.type == 'PhaseFieldCrystal2DSquare':
            string += "A: " + str(self.A) + "\n"
            string += "B: " + str(self.B) + "\n"
        elif self.type == 'PhaseFieldCrystal3DBodyCenteredCubic':
            string += "A: " + str(self.A) + "\n"
        elif self.type == 'PhaseFieldCrystal3DFaceCenteredCubic':
            string += "A: " + str(self.A) + "\n"
            string += "B: " + str(self.B) + "\n"
        elif self.type == 'PhaseFieldCrystal3DSimpleCubic':
            string += "A: " + str(self.A) + "\n"
            string += "B: " + str(self.B) + "\n"
            string += "C: " + str(self.C) + "\n"
        string += "-------------------------------\n"

        return string
    
    #######################################################
    ############ CONFIGURATION FUNCTIONS #################
    #######################################################

    def conf_PFC_from_amplitudes(self, eta=None, rotation=None):
        """Configures the PFC from the amplitudes.

        Args:
            eta: The amplitudes to configure the PFC from.

        Returns:
            Configures self.psi and self.psi_f.
        """

        self.psi = self.calc_PFC_from_amplitudes(eta, rotation)
        self.psi_f = self.fft(self.psi)

    def conf_advect_PFC(self,u):
        """Advects the PFC according to the displacement field u.

        Args:
            u: The displacement field to advect the PFC with.

        Returns:
            None, but updates the PFC.
        """

        self.psi = np.real(self.calc_advect_field(self.psi, u, self.psi_f))
        self.psi_f = self.fft(self.psi)

    def conf_apply_distortion(self, distortion, update_q_and_a_vectors=False):
        """Applies a distortion to the PFC.

        Args:
            distortion: The distortion to apply to the PFC.
        
        Returns:
            None, but updates the PFC.
        """

        if self.dim == 1:
            self.k[0] = self.k[0]/(1+distortion)
            self.dif[0] = self.dif[0]/(1+distortion)
            self.x = self.x*(1+distortion)
            self.xmax = self.xmax*(1+distortion)
            self.xmin = self.xmin*(1+distortion)
            self.size_x = self.size_x*(1+distortion)
            self.dx = self.dx*(1+distortion)
            self.volume = self.volume*(1+distortion)

        elif self.dim == 2:
            
            distortion_is_shear = False
            if isinstance(distortion, float):
                distortion_matrix = np.array([[1+distortion,0],[0,1+distortion]])
            else:
                distortion_matrix = np.eye(2) + np.array(distortion)

                if abs(distortion_matrix[0,1]) + abs(distortion_matrix[1,0])> 0:
                    distortion_is_shear = True

            # Updating the k and dif vectors
            inverse_distortion_matrix = np.linalg.inv(distortion_matrix)

            if distortion_is_shear:  
                self.bool_is_shear_distorted = True

                # Creating 2D meshgrid
                X,Y = np.meshgrid(self.x.flatten(),self.y.flatten(), indexing='ij')

                # Applying the strain
                X0 = X.copy()
                Y0 = Y.copy()
                X = distortion_matrix[0,0]*X0 + distortion_matrix[0,1]*Y0
                Y = distortion_matrix[1,0]*X0 + distortion_matrix[1,1]*Y0

                # Updating the x and y coordinates
                self.X = X
                self.Y = Y

                k00 = self.k[0].copy()
                k01 = self.k[1].copy()
                self.k[0] = k00*inverse_distortion_matrix[0,0] + k01*inverse_distortion_matrix[0,1]
                self.k[1] = k00*inverse_distortion_matrix[1,0] + k01*inverse_distortion_matrix[1,1]

            else:
                
                x0 = self.x.copy()
                y0 = self.y.copy()
                self.x = x0*distortion_matrix[0,0]
                self.y = y0*distortion_matrix[1,1] 

                X = self.x
                Y = self.y

                k00 = self.k[0].copy()
                k01 = self.k[1].copy()
                self.k[0] = k00*inverse_distortion_matrix[0,0]
                self.k[1] = k01*inverse_distortion_matrix[1,1]

            self.dif[0] = 1j*self.k[0]
            self.dif[1] = 1j*self.k[1]

            if update_q_and_a_vectors:
                    a00 = self.a[:,0].copy()
                    a01 = self.a[:,1].copy()
                    self.a[:,0] = a00*distortion_matrix[0,0] + a01*distortion_matrix[0,1]
                    self.a[:,1] = a00*distortion_matrix[1,0] + a01*distortion_matrix[1,1]

                    q00 = self.q[:,0].copy()
                    q01 = self.q[:,1].copy()
                    self.q[:,0] = q00*inverse_distortion_matrix[0,0] + q01*inverse_distortion_matrix[0,1]
                    self.q[:,1] = q00*inverse_distortion_matrix[1,0] + q01*inverse_distortion_matrix[1,1]

            # Updating the dx and dy
            original_dx = self.dx
            original_dy = self.dy

            self.dx = np.sqrt((distortion_matrix[0,0]*original_dx)**2 + (distortion_matrix[0,1]*original_dy)**2)
            self.dy = np.sqrt((distortion_matrix[1,0]*original_dx)**2 + (distortion_matrix[1,1]*original_dy)**2)

            # Updating the size_x and size_y
            original_size_x = self.size_x
            original_size_y = self.size_y
            self.size_x = np.sqrt((distortion_matrix[0,0]*original_size_x)**2 + (distortion_matrix[0,1]*original_size_y)**2)
            self.size_y = np.sqrt((distortion_matrix[1,0]*original_size_x)**2 + (distortion_matrix[1,1]*original_size_y)**2)

            # Updating the x and y coordinate limits
            self.xmax = np.max(X)
            self.xmin = np.min(X)

            self.ymax = np.max(Y)
            self.ymin = np.min(Y)

            volume_factor = np.linalg.det(distortion_matrix)
            self.dV = self.dV*volume_factor
            self.volume = self.volume*volume_factor


        elif self.dim == 3:
            # Distortion is shear
            distortion_is_shear = False
            if isinstance(distortion, float):
                distortion_matrix = np.array([[1+distortion,0,0],[0,1+distortion,0],[0,0,1+distortion]])
            else:
                distortion_matrix = np.eye(3) + np.array(distortion)

                if abs(distortion_matrix[0,1]) +\
                    abs(distortion_matrix[1,0]) + \
                    abs(distortion_matrix[0,2]) + \
                    abs(distortion_matrix[2,0]) + \
                    abs(distortion_matrix[1,2]) + \
                    abs(distortion_matrix[2,1]) > 0:
                    distortion_is_shear = True

            # Updating the k and dif vectors
            inverse_distortion_matrix = np.linalg.inv(distortion_matrix)

            # if distortion_is_shear:
            if distortion_is_shear:  
                self.bool_is_shear_distorted = True

                # Creating 2D meshgrid
                X,Y,Z = np.meshgrid(self.x.flatten(),self.y.flatten(), self.z.flatten(), indexing='ij')

                # Applying the strain
                X0 = X.copy()
                Y0 = Y.copy()
                Z0 = Z.copy()

                X = distortion_matrix[0,0]*X0 + distortion_matrix[0,1]*Y0 + distortion_matrix[0,2]*Z0
                Y = distortion_matrix[1,0]*X0 + distortion_matrix[1,1]*Y0 + distortion_matrix[1,2]*Z0
                Z = distortion_matrix[2,0]*X0 + distortion_matrix[2,1]*Y0 + distortion_matrix[2,2]*Z0

                # Updating the x and y coordinates
                self.X = X
                self.Y = Y
                self.Z = Z

                k00 = self.k[0].copy()
                k01 = self.k[1].copy()
                k02 = self.k[2].copy()
                self.k[0] = k00*inverse_distortion_matrix[0,0] + k01*inverse_distortion_matrix[0,1] + k02*inverse_distortion_matrix[0,2]
                self.k[1] = k00*inverse_distortion_matrix[1,0] + k01*inverse_distortion_matrix[1,1] + k02*inverse_distortion_matrix[1,2]
                self.k[2] = k00*inverse_distortion_matrix[2,0] + k01*inverse_distortion_matrix[2,1] + k02*inverse_distortion_matrix[2,2]

            else:
                
                x0 = self.x.copy()
                y0 = self.y.copy()
                z0 = self.z.copy()

                self.x = x0*distortion_matrix[0,0]
                self.y = y0*distortion_matrix[1,1] 
                self.z = z0*distortion_matrix[2,2]

                X = self.x
                Y = self.y
                Z = self.z

                k00 = self.k[0].copy()
                k01 = self.k[1].copy()
                k02 = self.k[2].copy()

                self.k[0] = k00*inverse_distortion_matrix[0,0]
                self.k[1] = k01*inverse_distortion_matrix[1,1]
                self.k[2] = k02*inverse_distortion_matrix[2,2]

            self.dif[0] = 1j*self.k[0]
            self.dif[1] = 1j*self.k[1]
            self.dif[2] = 1j*self.k[2]

            if update_q_and_a_vectors:
                    a00 = self.a[:,0].copy()
                    a01 = self.a[:,1].copy()
                    a02 = self.a[:,2].copy()
                    self.a[:,0] = a00*distortion_matrix[0,0] + a01*distortion_matrix[0,1] + a02*distortion_matrix[0,2]
                    self.a[:,1] = a00*distortion_matrix[1,0] + a01*distortion_matrix[1,1] + a02*distortion_matrix[1,2]
                    self.a[:,2] = a00*distortion_matrix[2,0] + a01*distortion_matrix[2,1] + a02*distortion_matrix[2,2]

                    q00 = self.q[:,0].copy()
                    q01 = self.q[:,1].copy()
                    q02 = self.q[:,2].copy()
                    self.q[:,0] = q00*inverse_distortion_matrix[0,0] + q01*inverse_distortion_matrix[0,1] + q02*inverse_distortion_matrix[0,2]
                    self.q[:,1] = q00*inverse_distortion_matrix[1,0] + q01*inverse_distortion_matrix[1,1] + q02*inverse_distortion_matrix[1,2]
                    self.q[:,2] = q00*inverse_distortion_matrix[2,0] + q01*inverse_distortion_matrix[2,1] + q02*inverse_distortion_matrix[2,2]

            # Updating the dx and dy
            original_dx = self.dx
            original_dy = self.dy
            original_dz = self.dz

            self.dx = np.sqrt((distortion_matrix[0,0]*original_dx)**2 + (distortion_matrix[0,1]*original_dy)**2 + (distortion_matrix[0,2]*original_dz)**2)
            self.dy = np.sqrt((distortion_matrix[1,0]*original_dx)**2 + (distortion_matrix[1,1]*original_dy)**2 + (distortion_matrix[1,2]*original_dz)**2)
            self.dz = np.sqrt((distortion_matrix[2,0]*original_dx)**2 + (distortion_matrix[2,1]*original_dy)**2 + (distortion_matrix[2,2]*original_dz)**2)

            # Updating the size_x and size_y
            original_size_x = self.size_x
            original_size_y = self.size_y
            original_size_z = self.size_z

            self.size_x = np.sqrt((distortion_matrix[0,0]*original_size_x)**2 + (distortion_matrix[0,1]*original_size_y)**2 + (distortion_matrix[0,2]*original_size_z)**2)
            self.size_y = np.sqrt((distortion_matrix[1,0]*original_size_x)**2 + (distortion_matrix[1,1]*original_size_y)**2 + (distortion_matrix[1,2]*original_size_z)**2)
            self.size_z = np.sqrt((distortion_matrix[2,0]*original_size_x)**2 + (distortion_matrix[2,1]*original_size_y)**2 + (distortion_matrix[2,2]*original_size_z)**2)

            # Updating the x and y coordinate limits
            self.xmax = np.max(X)
            self.xmin = np.min(X)

            self.ymax = np.max(Y)
            self.ymin = np.min(Y)

            self.zmax = np.max(Z)
            self.zmin = np.min(Z)

            volume_factor = np.linalg.det(distortion_matrix)
            self.dV = self.dV*volume_factor
            self.volume = self.volume*volume_factor
            

    def conf_strain_to_equilibrium(self):
        """Configures 
        Strains the pfc to equilibrium by adjusting the position variables and k-space variables.

        Args:
            None

        Returns:
            None, configures the PFC.
        """
        
        # Evolve PFC to reach initial state of equilibrium (100 time steps ok)
        number_of_initial_steps = 100
        self.evolve_PFC(number_of_initial_steps, suppress_output=True)
        number_of_steps = 10 # 10 steps is enough to reach equilibrium

        # Calculate average free energy to compare
        average_free_energy = self.calc_free_energy()/self.volume
        # print('Average free energy at initial state: ', average_free_energy)

        strain = 0
        strain_increment = 0.00001
        strain += strain_increment

        pfc_strained = copy.deepcopy(self)
        pfc_strained.conf_apply_distortion(strain, update_q_and_a_vectors=True)

        # Evolve PFC
        pfc_strained.evolve_PFC(number_of_steps, suppress_output=True)

        # Calculate free energy to compare
        average_free_energy_tmp = pfc_strained.calc_free_energy()/pfc_strained.volume

        positive_strain_required = False
        while average_free_energy_tmp < average_free_energy:

            positive_strain_required = True

            average_free_energy = average_free_energy_tmp

            strain += strain_increment

            pfc_strained = copy.deepcopy(self)
            pfc_strained.conf_apply_distortion(strain, update_q_and_a_vectors=True)
            
            pfc_strained.evolve_PFC(number_of_steps, suppress_output=True)

            average_free_energy_tmp = pfc_strained.calc_free_energy()/pfc_strained.volume
            # print('Average free energy at strain: ', strain, ' is: ', average_free_energy_tmp)
        
        if positive_strain_required:
            # Going one back to get the lowest free energy
            final_strain = strain - strain_increment
            self.conf_apply_distortion(final_strain, update_q_and_a_vectors=True)
            # tool_print_in_color('Lowest average free energy found at strain: ' + str(final_strain), 'green')

        else: #negative strain required

            strain = - strain_increment

            pfc_strained = copy.deepcopy(self)
            pfc_strained.conf_apply_distortion(strain, update_q_and_a_vectors=True)

            pfc_strained.evolve_PFC(number_of_steps, suppress_output=True)
            average_free_energy_tmp = pfc_strained.calc_free_energy()/pfc_strained.volume
            # print('Average Free energy at strain: ', strain, ' is: ', average_free_energy_tmp)

            while average_free_energy_tmp < average_free_energy:

                average_free_energy = average_free_energy_tmp

                strain -= strain_increment

                pfc_strained = copy.deepcopy(self)
                pfc_strained.conf_apply_distortion(strain, update_q_and_a_vectors=True)

                pfc_strained.evolve_PFC(number_of_steps, suppress_output=True)

                average_free_energy_tmp = pfc_strained.calc_free_energy()/pfc_strained.volume

                # print('Average free energy at strain: ', strain, ' is: ', average_free_energy_tmp)

            # Going one back to get the lowest free energy
            final_strain = strain + strain_increment
            self.conf_apply_distortion(final_strain, update_q_and_a_vectors=True)
            # tool_print_in_color('Lowest average free energy found at strain: ' + str(final_strain), 'green')   

        return final_strain

    def conf_create_polycrystal(self, type, **kwargs):
        """Creates a polycrystal.

        Args:
            type: The type of polycrystal to create.
            kwargs: Additional arguments for the polycrystal creation. Including:
                relaxation_time: The relaxation time to use for the polycrystal creation.
        
        Returns:
            None, but updates the PFC.
        """
        # First type of polycrystal, see documentation.
        if type == 'circular':
            # This creates a standard orientation of the crystal
            self.conf_PFC_from_amplitudes()

            rotation = kwargs.get('rotation', np.pi/6)
            position = kwargs.get('position', self.rmid)
            radius = kwargs.get('radius', self.size_min/4)

            # Create the rotated field
            psi_rotated = self.calc_PFC_from_amplitudes(rotation=[0,0,rotation])

            # Set the rotated field in the inclusion region
            region  = self.calc_region_disk(position, radius)
            self.psi[region] = psi_rotated[region]
            self.psi_f = self.fft(self.psi)

            # Smooth the interface
            relaxation_time = kwargs.get('relaxation_time', 10)
            self.evolve_PFC(round(relaxation_time/self.dt))

        elif type == 'four_grain':
            if self.dim == 1:
                raise Exception("Polycrystal type four_grain is not valid for 1 dimension.") 
                
            self.psi = self.calc_PFC_from_amplitudes(self.eta0)
            
            l1  = self.y>1/6*self.ymax+(4/6*self.ymax)/(1/1*self.xmax)*self.x
            l2  = self.y>1/2*self.ymax+(1/2*self.ymax)/(2/3*self.xmax)*self.x
            l3  = self.y>1/2*self.ymax-(1/2*self.ymax)/(2/3*self.xmax)*self.x
            l4  = self.y>5/6*self.ymax-(4/6*self.ymax)/(1/1*self.xmax)*self.x
            l5  = self.x>self.xmax/4
            l6  = self.y>5/4*self.ymax-(3/4*self.ymax)/(1/1*self.xmax)*self.x
            l7  = self.x>self.xmax/2
            l8  = self.y<1/2*self.ymax+(1/2*self.ymax)/(2/3*self.xmax)*self.x
            l9  = self.x>3/4*self.xmax
            l10 = self.y<-1/4*self.ymax+(3/4*self.ymax)/(1/1*self.xmax)*self.x

            zdir = 1 if self.dim == 2 else np.ones((1,1,self.zRes))

            pfcRotated = self.calc_PFC_from_amplitudes(self.eta0, rotation=[0,0,-22.5/180*np.pi])
            pfcRotated = np.roll(pfcRotated, -round(self.yRes/2), axis=1)
            pfcRotated = np.roll(pfcRotated, -round(self.xRes/3), axis=0)
            region = np.bool_((l4*~(l7)*~(l8) + ~(l1)*~(l3)*~(l7))*zdir)
            self.psi[region] = pfcRotated[region]

            pfcRotated = self.calc_PFC_from_amplitudes(self.eta0, rotation=[0,0,22.5/180*np.pi])
            pfcRotated = np.roll(pfcRotated, -round(self.yRes/2), axis=1)
            pfcRotated = np.roll(pfcRotated, round(self.xRes/3), axis=0)
            region = np.bool_((l1*l6*l7 + l7*l10*~(l4))*zdir)
            self.psi[region] = pfcRotated[region]

            pfcRotated = self.calc_PFC_from_amplitudes(self.eta0, rotation=[0,0,45/180*np.pi])
            pfcRotated = np.roll(pfcRotated, -round(self.xRes/2), axis=0)
            region = np.bool_((l1*~(l4)*~(l5) + ~(l1)*l4*l9)*zdir)
            self.psi[region] = pfcRotated[region]

            self.psi_f = self.fft(self.psi)

            relaxation_time = kwargs.get('relaxation_time', 10)
            
            self.evolve_PFC(round(relaxation_time/self.dt))


    #######################################################
    #######################################################
    ############### EVOLUTION FUNCTIONS ###################
    #######################################################
    #######################################################


    #######################################################
    #################### STANDARD #########################
    #######################################################
    def evolve_PFC(self, number_of_steps, method='ETD2RK', suppress_output=False):
        """Evolves the PFC according to classical PFC dynamics.

        Args:
            number_of_steps: The number of steps to evolve the PFC.
            method: The method to use for the evolution (default: ETD2RK).
        
        Returns:
            Updates self.psi and self.psi_f
        """

        if self.type_of_evolution == 'conserved':
            omega_f = -self.calc_k2()*(self.r + self.calc_L_f()**2)
            non_linear_evolution_function_f = self.calc_nonlinear_evolution_function_conserved_f
        elif self.type_of_evolution == 'unconserved':
            omega_f = -(self.r + self.calc_L_f()**2)
            non_linear_evolution_function_f = self.calc_nonlinear_evolution_function_unconserved_f

        integrating_factors_f, solver = self.calc_integrating_factors_f_and_solver(omega_f, method)

        for n in tqdm(range(number_of_steps), desc='Evolving the PFC (conserved)', disable=suppress_output):
            self.psi, self.psi_f = solver(integrating_factors_f,
                                          non_linear_evolution_function_f,
                                          self.psi, self.psi_f)

            # These steps seem to be necessary for numerical stability (Vidar 18.12.23)
            self.psi = np.real(self.psi)
            self.psi_f = self.fft(self.psi)

    # Nonlinear part conserved
    def calc_nonlinear_evolution_function_conserved_f(self, psi, t):
        return -self.calc_k2()*self.fft(self.t * psi ** 2 + self.v * psi ** 3)

    # Non-linear part unconserved
    def calc_nonlinear_evolution_function_unconserved_f(self, psi, t):
        return -self.fft(self.t * psi ** 2 + self.v * psi ** 3)
    #######################################################

    #######################################################
    ##### MECHANICAL EQUILIBRIUM (CONSERVED) ##############
    #######################################################
    def evolve_PFC_mechanical_equilibrium(self, time, Delta_t = 10, method='ETD2RK'):
        """Evolves the PFC in mechanical equilibrium. 

        Args:
            number_of_steps: The number of steps to evolve the PFC
            Delta_t: The time step for the mechanical equilibrium evolution
            method: The method to use for the evolution

        Returns:
            Updates self.psi and self.psi_f
        """
        
        number_of_steps = round(time/self.dt)
        number_of_steps_per_iteration = round(Delta_t/self.dt)
        number_of_iterations = round(number_of_steps/number_of_steps_per_iteration)

        for n in tqdm(range(number_of_iterations), desc='Evolving the PFC in mechanical equilibrium'):
            self.conf_advect_PFC(self.calc_displacement_field_to_equilibrium())
            self.evolve_PFC(number_of_steps_per_iteration, method)
    #######################################################

    #######################################################
    ######### HYDRODYNAMIC (CONSERVED) ####################
    #######################################################
    def evolve_PFC_hydrodynamic(self, number_of_steps, 
                                method = 'ETD2RK',
                                gamma_S = 2**-6,
                                rho0 = 2**-6):
        """Evolves the PFC according to hydrodynamic PFC dynamics.

        This requires introducing a velocity field. 
        If psi does not contain this field, it is added to the components psi[1],psi[2],psi[3].
        
        Args:
            number_of_steps: The number of steps to evolve the PFC.
            method: The method to use for the evolution (default: ETD2RK).
            gamma_S: The surface tension coefficient.
            rho0: The mass density.

        Returns:
            Updates self.psi and self.psi_f
        """
        
        if hasattr(self,'bool_has_velocity_field'):
            pass
        else:
            self.bool_has_velocity_field = True
            self.psi = np.array([self.psi]+[np.zeros_like(self.psi)]*self.dim)
            self.psi_f = np.array([self.psi_f]+[np.zeros_like(self.psi_f)]*self.dim)
            # print("psi shape", self.psi.shape)

            if not hasattr(self,'external_force_density_f'):
                self.external_force_density_f = np.zeros([self.dim] + self.dims, dtype=complex)

        self.gamma_S = gamma_S
        self.rho0 = rho0

        omega_f = self.calc_omega_hydrodynamic_f()
        integrating_factors_f, solver = self.calc_integrating_factors_f_and_solver(omega_f, method)

        for n in tqdm(range(number_of_steps), desc='Evolving the hydrodynamic PFC'):
            self.psi, self.psi_f = solver(integrating_factors_f,
                                          self.calc_nonlinear_hydrodynamic_evolution_function_f,
                                          self.psi, self.psi_f)
            
            self.psi = np.real(self.psi)
            self.psi_f = self.fft(self.psi )

    # Linear part
    def calc_omega_hydrodynamic_f(self):
        """Calculates the hydrodynamic evolution function omega_f.

        Args:
            None

        Returns:
            The hydrodynamic evolution function omega_f.
        """
        k2 = self.calc_k2()

        if self.type_of_evolution == 'conserved':
            omega_f = -self.calc_k2()*(self.r + self.calc_L_f()**2)
        elif self.type_of_evolution == 'unconserved':
            omega_f = -(self.r + self.calc_L_f()**2)

        return np.array([omega_f]+[-self.gamma_S/self.rho0*k2]*self.dim)

    # Nonlinear part
    def calc_nonlinear_hydrodynamic_evolution_function_f(self, field, t):
        """Calculates the hydrodynamic evolution function of the PFC.

        Args:
            field: The field to calculate the evolution function of.
            t: The time.

        Returns:
            The nonlinear evolution function for the hydrodynamic PFC.
        """

        field_f = self.fft(field)

        k2 = self.calc_k2()

        N0_f = -k2*self.fft(self.t * field[0] ** 2 + self.v * field[0] ** 3) \
            - self.fft(sum([field[i+1]*self.ifft(self.dif[i]*field_f[0]) for i in range(self.dim)]))
        
        force_density_f = self.calc_stress_divergence_f(field_f[0])

        return np.array([N0_f] + [1/self.rho0*(force_density_f[i]+self.external_force_density_f[i]) for i in range(self.dim)])
    #######################################################

    #######################################################
    #######################################################
    ############## CALCULATION FUNCTIONS ##################
    #######################################################
    #######################################################
    
    def calc_strained_amplitudes(self):
        """ Straines the PFC to equilibrium and returns the amplitudes.
    
        Args:
            None

        Returns:
            The amplitudes of the strained PFC.
        """
        # tool_print_in_color('Proto amplitudes and elastic constants', 'blue')
        # tool_print_in_color('---', 'blue')
        # print(f'Proto psi0: {self.psi0:.02f}')
        # # print(f'Proto A: {self.A:.05f}')
        # if self.type in ['PhaseFieldCrystal2DSquare','PhaseFieldCrystal3DFaceCenteredCubic','PhaseFieldCrystal3DSimpleCubic']:
        #     print(f'Proto B: {self.B:.05f}')
        # if self.type in ['PhaseFieldCrystal3DSimpleCubic']:
        #     print(f'Proto C: {self.C:.05f}')

        # print(f'Proto mu: {self.el_mu:.05f}')
        # print(f'Proto lambda: {self.el_lambda:.05f}')
        # print(f'Proto gamma: {self.el_gamma:.05f}')

        print('---')
        tool_print_in_color('Straining PFC to reach equilibrium', 'blue')
        print('---')

        # Strain PFC to equilibrium
        self.conf_PFC_from_amplitudes()
        final_strain = self.conf_strain_to_equilibrium()
        self.evolve_PFC(200, suppress_output=True)

        # tool_print_in_color('Amplitudes after strain', 'blue')
        # tool_print_in_color('---', 'blue')
        print(f'Equilibrium strain: {final_strain:.05f}')
        print(f'Equilibrium q-vector: {1/(1+final_strain):.05f}')

        psi0 = np.mean(self.psi)
        # print(f'Eq. psi0: {psi0:.02f}')

        eta = self.calc_demodulate_PFC(only_primary_modes=False)
        A = np.mean(np.real(eta[0]))
        number_of_independent_amplitudes = 1

        if self.type == 'PhaseFieldCrystal2DSquare':
            number_of_independent_amplitudes = 2
            B = np.mean(np.real(eta[2]))
        
        if self.type == 'PhaseFieldCrystal3DFaceCenteredCubic':
            number_of_independent_amplitudes = 2
            B = np.mean(np.real(eta[4]))

        if self.type == 'PhaseFieldCrystal3DSimpleCubic':
            number_of_independent_amplitudes = 3
            B = np.mean(np.real(eta[3]))
            C = np.mean(np.real(eta[9]))

        print('---')
        tool_print_in_color('Equilibrium amplitudes:', 'blue')
        print('---')
        print(f'Proto mean density psi0      : {self.psi0:.02f}')
        print(f'Equilibrium mean density psi0: {psi0:.02f}')
        print('Ratio (equilibrium/proto)     : {:.05f}'.format(psi0/self.psi0))

        if number_of_independent_amplitudes >= 1:
            print('---')
            print(f'Proto amplitude       A : {self.A:.05f}')
            print(f'Equilibrium amplitude A : {A:.05f}')
            print('Ratio (equilibrium/proto): {:.05f}'.format(A/self.A))
        if number_of_independent_amplitudes >= 2:
            print('---')
            print(f'Proto amplitude       B : {self.B:.05f}')
            print(f'Equilibrium amplitude B : {B:.05f}')
            print('Ratio (equilibrium/proto): {:.05f}'.format(B/self.B))
        
        if number_of_independent_amplitudes >= 3:
            print('---')
            print(f'Proto amplitude       C : {self.C:.05f}')
            print(f'Equilibrium amplitude C : {C:.05f}')
            print('Ratio (equilibrium/proto): {:.05f}'.format(C/self.C))
        
        # Finding elastic constants
        # def elastic_energy(strain, el_lambda, el_mu, el_gamma):
        #     exx, exy, eyy = strain
        #     return el_lambda/2*(exx+eyy)**2 + el_mu*(exx**2 + 2*exy**2 + eyy**2) + el_gamma/2*(exx**2 + eyy**2)

        # strain_magnitudes = np.linspace(-0.01,0.01,21)
        # exx = np.array([[a,0,a] for a in strain_magnitudes]).flatten()
        # exy = np.array([[0,a,0] for a in strain_magnitudes]).flatten()
        # eyy = np.array([[a,0,-a] for a in strain_magnitudes]).flatten()

        f0 = self.calc_free_energy()/self.volume

        # free_energies = np.zeros_like(exx)
        # for n in range(len(exx)):
        #     if self.dim == 2:
        #         distortion = [[exx[n], exy[n]],[exy[n], eyy[n]]]
        #     elif self.dim == 3:
        #         distortion = [[exx[n], exy[n], 0],[exy[n], eyy[n], 0],[0,0,0]]

        #     pfc_strained = copy.deepcopy(self)
        #     pfc_strained.conf_apply_distortion(distortion)
        #     f = pfc_strained.calc_free_energy()/pfc_strained.volume
        #     free_energies[n] = f-f0

        # params,_ = curve_fit(elastic_energy, (exx, exy, eyy), free_energies)
        # el_lambda, el_mu, el_gamma = params

        epsilon=0.01
        
        if self.dim == 2:
            pure_shear_distortion = [[0,epsilon],[epsilon,0]]
            volume_conserving_compression = [[epsilon,0],[0,-epsilon]]
            pure_compression_pos = [[epsilon,0],[0,epsilon]]
            pure_compression_neg = [[-epsilon,0],[0,-epsilon]]
        else:
            pure_shear_distortion = [[0,epsilon,0],[epsilon,0,0],[0,0,0]]
            volume_conserving_compression = [[epsilon,0,0],[0,-epsilon,0],[0,0,0]]
            pure_compression_pos = [[epsilon,0,0],[0,epsilon,0],[0,0,0]]
            pure_compression_neg = [[-epsilon,0,0],[0,-epsilon,0],[0,0,0]]

        pfc_strained = copy.deepcopy(self)
        pfc_strained.conf_apply_distortion(pure_shear_distortion)
        f_pure_shear = pfc_strained.calc_free_energy()/pfc_strained.volume - f0

        pfc_strained = copy.deepcopy(self)
        pfc_strained.conf_apply_distortion(volume_conserving_compression)
        f_volume_conserving_compression = pfc_strained.calc_free_energy()/pfc_strained.volume - f0

        pfc_strained = copy.deepcopy(self)
        pfc_strained.conf_apply_distortion(pure_compression_pos)
        f_pure_compression_pos = pfc_strained.calc_free_energy()/pfc_strained.volume - f0

        pfc_strained = copy.deepcopy(self)
        pfc_strained.conf_apply_distortion(pure_compression_neg)
        f_pure_compression_neg = pfc_strained.calc_free_energy()/pfc_strained.volume - f0

        f_pure_compression = (f_pure_compression_pos + f_pure_compression_neg)/2

        el_lambda = (0.5*f_pure_compression - 0.5*f_volume_conserving_compression)/(epsilon**2)
        el_mu = (0.5*f_pure_shear)/(epsilon**2)
        el_gamma = (-f_pure_shear + f_volume_conserving_compression)/(epsilon**2)


        if self.type == 'PhaseFieldCrystal1DPeriodic':
            el_lambda_from_eq_amplitudes = 2*(A**2)
        elif self.type == 'PhaseFieldCrystal2DTriangular':
            el_lambda_from_eq_amplitudes = 3*(A**2)
            el_mu_from_eq_amplitudes = 3*(A**2)
            el_gamma_from_eq_amplitudes = 0
        elif self.type == 'PhaseFieldCrystal2DSquare':
            el_lambda_from_eq_amplitudes = 16*(B**2)
            el_mu_from_eq_amplitudes = 16*(B**2)
            el_gamma_from_eq_amplitudes = 8*A**2 - 32*B**2
        elif self.type == 'PhaseFieldCrystal3DBodyCenteredCubic':
            el_lambda_from_eq_amplitudes = 4*(A**2)
            el_mu_from_eq_amplitudes = 4*(A**2)
            el_gamma_from_eq_amplitudes = -4*(A**2)
        elif self.type == 'PhaseFieldCrystal3DFaceCenteredCubic':
            el_lambda_from_eq_amplitudes = 32/81*(A**2)
            el_mu_from_eq_amplitudes = 32/81*(A**2)
            el_gamma_from_eq_amplitudes = 32/81*(2*B**2-A**2)
        elif self.type == 'PhaseFieldCrystal3DSimpleCubic':
            el_lambda_from_eq_amplitudes = 16*(B**2) + 128*(C**2)
            el_mu_from_eq_amplitudes = 16*(B**2) + 128*(C**2)
            el_gamma_from_eq_amplitudes = 32*(A**2) - 16*(B**2) - 256*(C**2)

        print('---')
        tool_print_in_color('Equilibrium elastic constants:', 'blue')
        print('---')
        print('Proto lambda                             : {:.05f}'.format(self.el_lambda))
        print('Equilibrium lambda (numerical)           : {:.05f}'.format(el_lambda))
        print('Equilibrium lambda (from eq. amplitudes) : {:.05f}'.format(el_lambda_from_eq_amplitudes))
        print('Ratio (equilibrium/proto)                : {:.05f}'.format(el_lambda/self.el_lambda))

        print('---')
        print('Proto mu                             : {:.05f}'.format(self.el_mu))
        print('Equilibrium mu (numerical)           : {:.05f}'.format(el_mu))
        print('Equilibrium mu (from eq. amplitudes) : {:.05f}'.format(el_mu_from_eq_amplitudes))
        print('Ratio (equilibrium/proto)            : {:.05f}'.format(el_mu/self.el_mu))

        print('---')
        print('Proto gamma                             : {:.05f}'.format(self.el_gamma))
        print('Equilibrium gamma (numerical)           : {:.05f}'.format(el_gamma))
        print('Equilibrium gamma (from eq. amplitudes) : {:.05f}'.format(el_gamma_from_eq_amplitudes))
        print('Ratio (equilibrium/proto)               : {:.05f}'.format(el_gamma/self.el_gamma))

        if number_of_independent_amplitudes == 1:
            return final_strain, psi0, A, el_lambda, el_mu, el_gamma
        elif number_of_independent_amplitudes == 2:
            return final_strain, psi0, A, B, el_lambda, el_mu, el_gamma
        elif number_of_independent_amplitudes == 3:
            return final_strain, psi0, A, B, C, el_lambda, el_mu, el_gamma

    def calc_PFC_free_energy_density_and_chemical_potential(self,field=None,field_f=None):
        """Calculates the free energy density and chemical potential of the PFC.

        Args:
            field: The field to calculate the free energy density and chemical potential of.
            field_f: The Fourier transform of the field.

        Returns:
            The free energy density and chemical potential of the PFC.
        """

        if field is None:
            field = self.psi
            field_f = self.psi_f

        psi_f = field_f
        
        # print("field shape",field.shape)
        # print("field_f shape",field_f.shape)

        psi = field 
        psi2 = field**2 
        psi3 = psi2*psi
        psi4 = psi2**2

        Lpsi = self.ifft(self.calc_L_f()*psi_f)

        free_energy_density = 1/2*Lpsi**2 \
            + 1/2*self.r*psi2 + 1/3*self.t*psi3 + 1/4*self.v*psi4
        
        # print("Lpsi shape",self.calc_Lpsi(psi_f).shape)

        L2psi = self.ifft(self.calc_L_f()**2*psi_f)

        chemical_potential =L2psi*psi \
            + self.r*psi + self.t*psi2 + self.v*psi3
        
        # print("L2psi shape",self.calc_L2psi(psi_f).shape)
        
        # print("Free energy density shape", free_energy_density.shape)
        # print("Chem pot shape", chemical_potential.shape)
        return np.real(free_energy_density), np.real(chemical_potential)
    

    def calc_displacement_field_to_equilibrium(self):
        """Calculates the displacement field needed to put the PFC in mechanical equilibrium.

        Args:
            None
        
        Returns:
            The displacement field u
        """

        # Calculate the stress divergence
        g_f = self.calc_stress_divergence_f()

        # Calculate the displacement field
        u_f = np.zeros([self.dim] + self.dims, dtype = complex)

        k2 = self.calc_k2()
        k = np.sqrt(k2)

        kappa = np.zeros([self.dim] + self.dims)
        for l in range(self.dim):
            kappa[l] = self.k[l]/k

        second_term_factor = (self.el_mu + self.el_lambda)/(1+sum([(self.el_mu + self.el_lambda)/(self.el_mu + self.el_gamma*kappa[l]**2)*kappa[l]**2 for l in range(self.dim)]))

        # print(second_term_factor)

        for i in range(self.dim):
            for j in range(self.dim):
                #Calculate the Green's function
                Gij_f = int(i==j)/(self.el_mu + self.el_gamma*kappa[i]**2) \
                - kappa[i]*kappa[j]/((self.el_mu + self.el_gamma*kappa[i]**2)*(self.el_mu + self.el_gamma*kappa[j]**2))*second_term_factor

                u_f[i] += 1/k2*Gij_f*g_f[j]
            
            # Set the zero mode to zero
            u_f[i][self.zero_index] = 0

            # print(u_f[i])

        return np.real(self.ifft(u_f))

        
    # Initial configuration methods
    def calc_amplitudes_with_dislocation(self, eta=None, x=None, y=None, dislocation_type=1):
        """ Calculates the amplitudes with a single point dislocation inserted.

        Args:
            eta: The amplitudes to insert the dislocation in.
            x: The x-coordinate of the dislocation.
            y: The y-coordinate of the dislocation.
            dislocation_type: The dislocation type to insert.
        
        Returns:
            The amplitudes containing the dislocation
        """

        if not (self.dim == 2):
            raise Exception("The dimension of the system must be 2 for a single point dislocation.")

        if x == None:
            x = self.xmid
        if y == None:
            y = self.ymid
        if eta == None:
            eta = self.eta0

        sn = self.dislocation_charges[dislocation_type - 1]
        for n in range(self.number_of_reciprocal_lattice_modes):
            if sn[n] != 0:
                eta[n] *= np.exp(1j * self.calc_angle_field_single_vortex([x, y], charge=sn[n]))

        return eta

    def calc_amplitudes_with_dislocation_dipole(self, eta=None,
                                                x1=None, y1=None,
                                                x2=None, y2=None,
                                                dislocation_type=1):
        """
        Inserts a  dislocation dipole in the system corresponding to dislocation type and its negative

        Args:
            eta: The amplitudes to insert the dislocation dipole in
            x1: The x-coordinate of the first dislocation
            y1: The y-coordinate of the first dislocation
            x2: The x-coordinate of the second dislocation
            y2: The y-coordinate of the second dislocation
            dislocation_type: The dislocation type to insert
        
        Returns:
            The amplitudes with the dislocation dipole inserted
        """

        if not (self.dim == 2):
            raise Exception("The dimension of the system must be 2 to insert a dislocation dipole.")

        if x1 == None:
            x1 = self.xmax / 3
        if y1 == None:
            y1 = self.ymax / 2
        if x2 == None:
            x2 = 2 * self.xmax / 3
        if y2 == None:
            y2 = self.ymax / 2

        if eta is None:
            eta = self.eta0.reshape(self.number_of_reciprocal_lattice_modes,1,1)\
                *np.ones([self.number_of_reciprocal_lattice_modes] + self.dims, dtype=complex)

        sn = self.dislocation_charges[dislocation_type - 1]

        theta = self.calc_angle_field_vortex_dipole(
            dipole_vector=[x2 - x1, y2 - y1],
            dipole_position=[(x1 + x2) / 2, (y1 + y2) / 2])

        for n in range(self.number_of_reciprocal_lattice_modes):
            if sn[n] != 0:
                eta[n] *= np.exp(1j * sn[n] * theta)

        return eta

    def calc_amplitudes_with_dislocation_ring(self, eta=None,
                                                position=None,
                                                radius=None,
                                                normal_vector=[0, 0, 1],
                                                dislocation_type=1):
        """
        Inserts a  dislocation ring in the system corresponding to dislocation type.

        Args:
            eta: The amplitudes to insert the dislocation ring in
            position: The position of the dislocation ring
            radius: The radius of the dislocation ring
            normal_vector: The normal vector of the dislocation ring
            dislocation_type: The dislocation type to insert

        Returns:
            The amplitudes with the dislocation ring inserted

        """

        if not (self.dim == 3):
            raise Exception("The dimension of the system must be 3 to insert a dislocation dipole.")

        if position == None:
            position = self.rmid

        if radius == None:
            radius = min(self.rmax)/3

        if eta == None:
            eta = self.eta0.reshape(self.number_of_reciprocal_lattice_modes,1,1,1)\
                *np.ones([self.number_of_reciprocal_lattice_modes] + self.dims, dtype=complex)

        sn = self.dislocation_charges[dislocation_type - 1]

        theta = self.calc_angle_field_vortex_ring(
            radius=radius,
            position=position,
            normal_vector=normal_vector)

        for n in range(self.number_of_reciprocal_lattice_modes):
            if sn[n] != 0:
                eta[n] *= np.exp(1j * sn[n] * theta)

        return eta
    
    def calc_PFC_from_amplitudes(self, eta=None, rotation=None):
        """Calculates the PFC from the amplitudes.

        Args:
            eta: The amplitudes to calculate the PFC from.
            rotation : The rotation of the PFC

        Returns:
            The PFC.
        """ 

        psi = self.psi0

        if rotation is None:
            rotation = [0,0,0]
        else:
            if isinstance(rotation, float):
                rotation = [0,0,rotation]

        rotation = sp.spatial.transform.Rotation.from_rotvec(rotation)
        rotation_matrix = rotation.as_matrix()

        # Rotate q-vectors to the new orientation
        if self.dim == 1:
            q = self.q # No rotation in 1D
        if self.dim == 2:
            q = (rotation_matrix[0:2,0:2]@self.q.transpose()).transpose()
        elif self.dim == 3:
            q = (rotation_matrix[0:3,0:3]@self.q.transpose()).transpose()

        if eta is None:
            eta = self.eta0

        for n in range(self.number_of_reciprocal_lattice_modes):

            if self.dim == 1:
                psi += 2 * eta[n] * np.exp(1j * q[n][0] * self.x)

            elif self.dim == 2:
                psi += 2 * eta[n] * np.exp(1j * (q[n][0] * self.x + q[n][1] * self.y))

            elif self.dim == 3:
                psi += 2 * eta[n] * np.exp(1j * (q[n][0] * self.x + q[n][1] * self.y + q[n][2] * self.z))

        return np.real(psi)

    def calc_demodulate_PFC(self, only_primary_modes=True):
        """Demodulates the PFC.

        Args:
            None

        Returns:
            The amplitudes corresponding to the demodulated PFC.
        """
        
        number_of_modes = self.number_of_primary_reciprocal_lattice_modes if only_primary_modes else self.number_of_reciprocal_lattice_modes

        eta = np.zeros([number_of_modes] + self.dims, 
                       dtype=complex)

        Gaussian_filter_f = self.calc_Gaussian_filter_f()

        order_parameter = self.psi if self.psi.ndim == self.dim else self.psi[0]

        if self.dim == 2:
                for n in range(number_of_modes):
                    eta[n] = self.ifft(Gaussian_filter_f*self.fft(order_parameter*np.exp(
                        -1j*self.q[n][0]*self.x - 1j*self.q[n][1]*self.y)))

        elif self.dim == 3:
            for n in range(number_of_modes):
                eta[n] = self.ifft(Gaussian_filter_f*self.fft(order_parameter*np.exp(
                    -1j*self.q[n][0]*self.x - 1j*self.q[n][1]*self.y - 1j*self.q[n][2]*self.z  
                    )))
                
        return eta

    def calc_stress_tensor_microscopic(self):
        """Calculates the microscopic stress of the phase-field crystal.

        Args:
            None

        Returns:
            The microscopic stress of the phase-field crystal.
        """
        if self.dim==1:
            raise Exception("The stress tensor is not yet defined in 1D.")
        elif self.dim==2:
            stress = np.zeros((3,self.xRes,self.yRes))

            Lpsi = np.real(self.ifft(self.calc_L_f()*self.psi_f))
            stress[0] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[0]*self.dif[0]*self.psi_f))
            stress[1] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[0]*self.dif[1]*self.psi_f))
            stress[2] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[1]*self.dif[1]*self.psi_f))
        
        elif self.dim==3:
            stress = np.zeros((6,self.xRes,self.yRes,self.zRes))

            Lpsi = np.real(self.ifft(self.calc_L_f()*self.psi_f))
            stress[0] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[0]*self.dif[0]*self.psi_f))
            stress[1] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[0]*self.dif[1]*self.psi_f))
            stress[2] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[0]*self.dif[2]*self.psi_f))
            stress[3] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[1]*self.dif[1]*self.psi_f))
            stress[4] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[1]*self.dif[2]*self.psi_f))
            stress[5] = -2*Lpsi*np.real(self.ifft(self.calc_L_sum_f()*self.dif[2]*self.dif[2]*self.psi_f))

        return stress

    def calc_stress_tensor(self):
        """Calculates the stress of the phase-field crystal.

        Args:
            None

        Returns:
            The stress tensor
        """
        
        stress = self.calc_stress_tensor_microscopic()

        Gaussian_filter_f = self.calc_Gaussian_filter_f()

        # Number of independent stress components
        if self.dim == 1:
            number_of_stress_components = 1
        elif self.dim == 2:
            number_of_stress_components = 3
        elif self.dim == 3:
            number_of_stress_components = 6

        # Coarse-grain the microscopic stress
        for n in range(number_of_stress_components):
            stress[n] = Gaussian_filter_f*self.fft(stress[n])

        # return stress
        # return np.real(self.ifft(Gaussian_filter_f))
        return np.real(self.ifft(stress))

    def calc_stress_divergence_f(self, field_f = None):
        """Calculates the divergence of the stress tensor in Fourier space.
        
        Args:
            field_f: The field in Fourier space.
        
        Returns:
            The divergence of the stress tensor in Fourier space.
        """
        if field_f is None:
            PFC_has_velocity_field = hasattr(self, 'bool_has_velocity_field') and self.bool_has_velocity_field
            if PFC_has_velocity_field:
                field_f = self.psi_f[0]
            else:
                field_f = self.psi_f

        L_f = self.calc_L_f()
        Lpsi = self.ifft(L_f*field_f)

        L_sum_f = self.calc_L_sum_f()
        k2 = self.calc_k2()

        return np.array([
            -2*self.calc_Gaussian_filter_f()*self.fft(
                sum([
                self.ifft(L_f*self.dif[i]*field_f)*self.ifft(L_sum_f*self.dif[i]*self.dif[j]*field_f) 
                for i in range(self.dim)
                ]) 
                +Lpsi*self.ifft(L_sum_f*self.dif[j]*(-k2)*field_f)) 
                for j in range(self.dim)]
                )


    def calc_structure_tensor_f(self):
        """Calculates the structure tensor of the phase-field crystal in Fourier space.

        Args:
            None

        Returns:
            The structure tensor in Fourier space.
        """

        if hasattr(self,'bool_has_velocity_field') and self.bool_has_velocity_field:
            field_f = self.psi_f[0]
        else:
            field_f = self.psi_f

        # Calculate the gradient
        diPsi = np.zeros([self.dim] + self.dims, dtype=complex)
        for i in range(self.dim):
            diPsi[i] = self.dif[i]*field_f
        diPsi = np.real(self.ifft(diPsi))

        if self.dim == 1:
            number_of_independent_strain_components = 1
        elif self.dim == 2:
            number_of_independent_strain_components = 3
        elif self.dim == 3:
            number_of_independent_strain_components = 6

        structure_tensor_f = np.zeros([number_of_independent_strain_components] + self.dims, dtype=complex)

        Gaussian_filter_f = self.calc_Gaussian_filter_f()
        if self.dim == 1:
            structure_tensor_f[0] = Gaussian_filter_f*self.fft(diPsi[0]*diPsi[0])
        elif self.dim == 2:
            structure_tensor_f[0] = Gaussian_filter_f*self.fft(diPsi[0]*diPsi[0])
            structure_tensor_f[1] = Gaussian_filter_f*self.fft(diPsi[0]*diPsi[1])
            structure_tensor_f[2] = Gaussian_filter_f*self.fft(diPsi[1]*diPsi[1])
        elif self.dim == 3:
            structure_tensor_f[0] = Gaussian_filter_f*self.fft(diPsi[0]*diPsi[0])
            structure_tensor_f[1] = Gaussian_filter_f*self.fft(diPsi[0]*diPsi[1])
            structure_tensor_f[2] = Gaussian_filter_f*self.fft(diPsi[0]*diPsi[2])
            structure_tensor_f[3] = Gaussian_filter_f*self.fft(diPsi[1]*diPsi[1])
            structure_tensor_f[4] = Gaussian_filter_f*self.fft(diPsi[1]*diPsi[2])
            structure_tensor_f[5] = Gaussian_filter_f*self.fft(diPsi[2]*diPsi[2])

        return structure_tensor_f
            

    def calc_structure_tensor(self):
        """Calculates the structure tensor of the phase-field crystal.

        Args:
            None
        Returns:
            The structure tensor.
        """
        structure_tensor_f = self.calc_structure_tensor_f()
        return np.real(self.ifft(structure_tensor_f))
        

    def calc_strain_tensor(self):
        """Calculates the strain of the phase-field crystal.

        Args:
            None

        Returns:
            The strain tensor
        """
        
        strain = -self.dim/(2*self.Phi)*self.calc_structure_tensor()

        if self.dim == 1:
            strain = 1/2 + strain
        elif self.dim == 2:
            strain[0] = 1/2 + strain[0]
            strain[2] = 1/2 + strain[2]
        elif self.dim == 3:
            strain[0] = 1/2 + strain[0]
            strain[3] = 1/2 + strain[3]
            strain[5] = 1/2 + strain[5]
        
        return strain
        



    def calc_dislocation_density(self, eta = None):
        """Calculates the dislocation density

        Args:
            eta: The amplitudes to calculate the dislocation density from.

        Returns:
            The dislocation density
        """

        if eta is None:
            eta = self.calc_demodulate_PFC()

        if self.dim == 2:
            alpha = np.zeros([2] + self.dims)

            for n in range(self.number_of_primary_reciprocal_lattice_modes):
                D = self.calc_determinant_field([np.real(eta[n]), np.imag(eta[n])])
                alpha[0] += D*self.q[n,0]
                alpha[1] += D*self.q[n,1]

        elif self.dim == 3:
            alpha = np.zeros([3,3] + self.dims)

            for n in range(self.number_of_primary_reciprocal_lattice_modes):
                D = self.calc_determinant_field([np.real(eta[n]), np.imag(eta[n])])
                for i in range(3):
                    alpha[i][0] += D[i]*self.q[n,0]
                    alpha[i][1] += D[i]*self.q[n,1]
                    alpha[i][2] += D[i]*self.q[n,2]

        alpha = 2*self.dim/(self.number_of_primary_reciprocal_lattice_modes*self.A**2)*alpha

        return alpha
        

    def calc_dislocation_nodes(self):
        """ Calculates the dislocation nodes

        Args:
            None
        
        Returns:
            The dislocation nodes
        """

        # At the moment (24.09.24 - Vidar), this function only works for unstrained PFCs
        if hasattr(self,'bool_is_shear_distorted') and self.bool_is_shear_distorted:
            raise Exception("Dislocation nodes cannot be calculated for distorted PFCs.")

        alpha = self.calc_dislocation_density()

        if self.dim == 2:
            # self.plot_field(np.sqrt(alpha[0]**2 + alpha[1]**2))
            # plt.show()
            # print(alpha)
            dislocation_nodes = self.calc_defect_nodes(np.sqrt(alpha[0]**2 + alpha[1]**2),
                                                       charge_tolerance=0.2*self.a0)
            
            for dislocation_node in dislocation_nodes:
                Burgers_vector = np.array([
                    alpha[0][dislocation_node['position_index']], 
                    alpha[1][dislocation_node['position_index']]
                    ])
                
                # print("Burgers vector:", Burgers_vector)
                
                # Find the Burgers vector 
                biggest_overlap = 0
                for a in self.a:
                    for sign in [-1, 1]:
                        overlap = np.dot(Burgers_vector, sign*a)
                        # print("Burgers vector",Burgers_vector)
                        # print("a-vector",a*sign)
                        # print("overlap",overlap)
                        # print("biggest overlap", biggest_overlap)
                        
                        if overlap > biggest_overlap:
                            # print("tjobing")
                            biggest_overlap = overlap
                            dislocation_node['Burgers_vector'] = sign*a
                
        elif self.dim == 3:
            dislocation_nodes = self.calc_defect_nodes(
                np.sqrt(
                      alpha[0][0]**2 + alpha[0][1]**2 + alpha[0][2]**2 \
                    + alpha[1][0]**2 + alpha[1][1]**2 + alpha[1][2]**2 \
                    + alpha[2][0]**2 + alpha[2][1]**2 + alpha[2][2]**2
                )
            )

            for dislocation_node in dislocation_nodes:
                alpha_tensor = np.array([
                    [alpha[0][0][dislocation_node['position_index']], alpha[0][1][dislocation_node['position_index']], alpha[0][2][dislocation_node['position_index']]],
                    [alpha[1][0][dislocation_node['position_index']], alpha[1][1][dislocation_node['position_index']], alpha[1][2][dislocation_node['position_index']]],
                    [alpha[2][0][dislocation_node['position_index']], alpha[2][1][dislocation_node['position_index']], alpha[2][2][dislocation_node['position_index']]]
                ])
                # print("alpha-tensor:",alpha_tensor)

                U, S, V = np.linalg.svd(alpha_tensor)
                tangent_vector = U[:,0]
                Burgers_vector = V[0,:]

                # print("U:",U)
                # print("S:",S)   
                # print("V:",V)

                # print("tangent vector:", tangent_vector)
                # print("Burgers vector:", Burgers_vector)

                # Find the Burgers vector 
                biggest_overlap = 0
                tangent_vector_sign = np.nan
                # print("Finding new Burgers vector...")
                for a in self.a:
                    for sign in [-1, 1]:
                        # print("Burgers vector",Burgers_vector)
                        # print("a-vector",a*sign)
                        # print("overlap",np.dot(Burgers_vector,sign*a))
                        overlap = np.dot(Burgers_vector, sign*a)
                        if overlap > biggest_overlap:
                            biggest_overlap = overlap
                            # print("New biggest overlap", overlap)
                            # print("Setting new BUrgers vector",a)
                            dislocation_node['Burgers_vector'] = a
                            tangent_vector_sign = sign

                dislocation_node['tangent_vector'] = tangent_vector_sign*tangent_vector

        return dislocation_nodes

    def calc_orientation_field(self):
        """Calculates the orientation field of the phase-field crystal.

        Args:
            None
        
        Returns:
            An orientation field, which is a vector field specifying the orientation of the crystal.
        """

        order_parameter = self.psi if self.psi.ndim == self.dim else self.psi[0]

        eta = np.zeros([self.number_of_primary_reciprocal_lattice_modes] + self.dims, 
                       dtype=complex)
        Gaussian_filter_f = self.calc_Gaussian_filter_f()

        if self.dim == 2:
            if self.type == 'PhaseFieldCrystal2DTriangular':
                wrap_factor = 3
            elif self.type == 'PhaseFieldCrystal2DSquare':
                wrap_factor = 2

            resolution = 16
            orientation_field = np.zeros((2,self.xRes,self.yRes))
            angles = np.arange(resolution)/resolution*np.pi/wrap_factor

            for angle in angles:
                rotation = sp.spatial.transform.Rotation.from_rotvec([0,0,angle])
                rotation_matrix = rotation.as_matrix()
                q = (rotation_matrix[0:2,0:2]@self.q.transpose()).transpose()
                for n in range(self.number_of_primary_reciprocal_lattice_modes):
                    eta[n] = self.ifft(Gaussian_filter_f*self.fft(order_parameter*np.exp(
                            -1j*q[n][0]*self.x - 1j*q[n][1]*self.y)))
                
                Phi2 = np.zeros(self.dims)
                for n in range(self.number_of_primary_reciprocal_lattice_modes):
                    Phi2 += abs(eta[n])**2
                
                orientation_field[0] += Phi2*np.cos(2*wrap_factor*angle)
                orientation_field[1] += Phi2*np.sin(2*wrap_factor*angle)

            return orientation_field
        
        elif self.dim == 3:

            # 1153 points
            xRes = 17
            yRes = 17
            zRes = 9

            x = np.linspace(-np.pi,np.pi, xRes)
            y = np.linspace(-np.pi,np.pi, yRes)
            z = np.linspace(0,np.pi, zRes)

            [X,Y,Z] = np.meshgrid(x,y,z,indexing='ij')

            # Flatten the meshgrid arrays into 1D arrays
            X_flat = X.flatten()
            Y_flat = Y.flatten()
            Z_flat = Z.flatten()

            # Compute the distance from the origin for each point
            distances = np.sqrt(X_flat**2 + Y_flat**2 + Z_flat**2)

            # Filter the points that lie inside the unit sphere
            inside_sphere = distances <= np.pi

            # Create a list of points that are inside the unit sphere
            points_inside_sphere = np.vstack((X_flat[inside_sphere], Y_flat[inside_sphere], Z_flat[inside_sphere])).T

            # Convert to a list of tuples
            points_inside_sphere_list = [tuple(point) for point in points_inside_sphere]

            # print(len(points_inside_sphere_list))
            # fig = plt.figure(figsize=(12, 12))
            # ax = fig.add_subplot(projection='3d')
            # ax.scatter(*zip(*points_inside_sphere_list))
            # plt.show()

            orientation_field = np.zeros((4,self.xRes,self.yRes,self.zRes))

            for point in points_inside_sphere_list:
                rotation = sp.spatial.transform.Rotation.from_rotvec(point)
                rotation_matrix = rotation.as_matrix()
                q = (rotation_matrix@self.q.transpose()).transpose()

                for n in range(self.number_of_primary_reciprocal_lattice_modes):
                    eta[n] = self.ifft(Gaussian_filter_f*self.fft(order_parameter*np.exp(
                            -1j*q[n][0]*self.x - 1j*q[n][1]*self.y)))
                
                Phi2 = np.zeros(self.dims)
                for n in range(self.number_of_primary_reciprocal_lattice_modes):
                    Phi2 += abs(eta[n])**2

                theta = np.sqrt(point[0]**2 + point[1]**2 + point[2]**2)
                nx = point[0]/theta if theta != 0 else 0
                ny = point[1]/theta if theta != 0 else 0
                nz = point[2]/theta if theta != 0 else 0

                orientation_field[0] += Phi2*np.cos(2*theta)
                orientation_field[1] += Phi2*np.sin(2*theta)*nx
                orientation_field[2] += Phi2*np.sin(2*theta)*ny
                orientation_field[3] += Phi2*np.sin(2*theta)*nz
            
            return orientation_field
                
    def calc_free_energy(self):
        free_energy_density, _ = self.calc_PFC_free_energy_density_and_chemical_potential()
        return self.calc_integrate_field(free_energy_density)

    #######################################################
    ############### PLOTTING FUNCTIONS ####################
    #######################################################

    def plot_field(self, field, **kwargs):
        """Plots the PFC

        Args:
            field: The field to plot.
            **kwargs: Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/ for a full list of keyword arguments.
        Returns:
            ax, fig: The axes and figure containing the plot.
        """
        
        PFC_is_distorted = True if hasattr(self, 'bool_is_shear_distorted') and self.bool_is_shear_distorted else False

        if PFC_is_distorted:
            kwargs['xmin'] = np.min(self.X)
            kwargs['xmax'] = np.max(self.X)
            kwargs['ymin'] = np.min(self.Y)
            kwargs['ymax'] = np.max(self.Y)
            kwargs['X'] = self.X
            kwargs['Y'] = self.Y
        
        if self.plot_lib == 'matplotlib':
            return plot_field_matplotlib(self, field, **kwargs)
        elif self.plot_lib == 'plotly':
            return plot_field_plotly(self, field, **kwargs)

    def plot_PFC(self, **kwargs):
        """Plots the PFC

        Args:
            **kwargs: Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/ for a full list of keyword arguments.
        Returns:
            ax, fig: The axes and figure containing the plot.
        """
        PFC_has_velocity_field = hasattr(self, 'bool_has_velocity_field') and self.bool_has_velocity_field
    
        kwargs['colormap'] = kwargs.get('colormap', 'Viridis')
        if PFC_has_velocity_field:
            return self.plot_field(self.psi[0], **kwargs)
        else:
            return self.plot_field(self.psi, **kwargs)



    def plot_orientation_field(self, orientation_field=None, **kwargs):
        """Plots the orientation field of the phase-field crystal.

        Args:
            orientation_field: The orientation field to plot.
            **kwargs: Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/ for a full list of keyword arguments.

        Returns:
            The axes containing the plot.
        """

        if orientation_field is None:
            orientation_field = self.calc_orientation_field()
        
        if self.dim == 2:
            complex_field = orientation_field[0] + 1j*orientation_field[1]
             
            if self.type == 'PhaseFieldCrystal2DTriangular':
                kwargs['cticks'] = [-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]
                kwargs['cticklabels'] = [r'$-\pi/6$', r'$-\pi/9$', r'$-\pi/18$', r'$0$', r'$\pi/18$', r'$\pi/9$', r'$\pi/6$']

            elif self.type == 'PhaseFieldCrystal2DSquare':
                kwargs['cticks'] = [-np.pi, -np.pi / 2, 0, np.pi / 2, np.pi]
                kwargs['cticklabels'] = [r'$-\pi/4$', r'$-\pi/8$', r'$0$', r'$\pi/8$', r'$\pi/4$']

            kwargs['plot_method'] = 'phase_angle'

            return self.plot_complex_field(complex_field, **kwargs)

        elif self.dim==3:
            theta = np.arccos(orientation_field[0])/2
            nnorm = np.sqrt(orientation_field[1]**2 + orientation_field[2]**2 + orientation_field[3]**2)
            nx = orientation_field[1]/nnorm
            ny = orientation_field[2]/nnorm 
            nz = orientation_field[3]/nnorm 

            vector_field = theta*np.array([nx,ny,nz])

            return self.plot_vector_field(vector_field, spacing=5, **kwargs)



from plotly.subplots import make_subplots

def plot_subplots_plotly(number_of_rows, number_of_columns, *args):
    """Makes subplots from a series of plotly figures.

    Args:
        *args: A series of plotly figures.
    Returns:
        A plotly figure with subplots.
    """
    if len(args) > number_of_rows * number_of_columns:
        raise Exception("There are too many subplots wrt. the number of rows and columns.")

    specs_list = [[{} for _ in range(number_of_columns)] for _ in range(number_of_rows)]

    # List of known 3D plot types in Plotly
    types_of_3D_traces = {'scatter3d', 'surface', 'mesh3d', 'cone', 'streamtube', 'isosurface', 'volume'}

    n = 0
    for row in specs_list:
        for spec in row:
            subplot_fig = args[n]
            if len(subplot_fig.data) > 1:
                print('\033[91m Warning: Multiple traces in the same subplot. This might cause issues. \033[0m')

            for trace in subplot_fig.data:  # Might be some issue with different types of traces in the same figure (Vidar 12.09.24)
                if trace.type in types_of_3D_traces:
                    spec['type'] = 'surface'
                else:
                    spec['type'] = 'xy'
            n += 1

    # Adjust horizontal_spacing and vertical_spacing to add more room between subplots
    fig = make_subplots(
        rows=number_of_rows, 
        cols=number_of_columns, 
        specs=specs_list,
        horizontal_spacing=0.1,  # Adjust this value as needed
        vertical_spacing=0.1     # Adjust this value as needed
    )

    n = 0
    n_2d = -1  # Counter for 2D plots
    n_3d = -1  # Counter for 3D plots
    for row in range(1, number_of_rows + 1):
        for col in range(1, number_of_columns + 1):
            subplot_fig = args[n]
            plot_is_3D = any(trace.type in types_of_3D_traces for trace in subplot_fig.data)

            if plot_is_3D:
                n_3d += 1
            else:
                n_2d += 1
                xaxis_id = f'x{n_2d+1}'
                yaxis_id = f'y{n_2d+1}'

            for trace in subplot_fig.data:
                trace_has_colorbar = hasattr(trace, 'colorbar') and trace.colorbar is not None
                marker_trace_has_colorbar = hasattr(trace, 'marker') and hasattr(trace.marker, 'colorbar') and trace.marker.colorbar is not None

                if trace_has_colorbar or marker_trace_has_colorbar:
                    colorbar = trace.colorbar if trace_has_colorbar else trace.marker.colorbar

                    colorbar.xanchor = 'left'
                    colorbar.x = (col / number_of_columns)**(9/8)

                    colorbar.yanchor = 'bottom'
                    colorbar.y = (number_of_rows - row) / number_of_rows

                    length = 1 / number_of_rows
                    colorbar.len = length

                if hasattr(trace, 'showlegend'):
                    trace.showlegend = False
                    
                fig.add_trace(trace, row=row, col=col)

            # Update x and y axes layout for each subplot (for 2D plots)
            if not plot_is_3D:
                fig.update_xaxes(subplot_fig.layout.xaxis, row=row, col=col)
                fig.update_yaxes(subplot_fig.layout.yaxis, row=row, col=col)

                if 'scaleanchor' in subplot_fig.layout.yaxis and subplot_fig.layout.yaxis.scaleanchor == 'x':
                    fig.update_yaxes(scaleanchor=xaxis_id, scaleratio=1, row=row, col=col)

                fig.update_xaxes(showgrid=True, row=row, col=col)
                fig.update_yaxes(showgrid=True, row=row, col=col)

            # Update 3D scene layout for each subplot (for 3D plots)
            if plot_is_3D:
                fig.update_scenes(subplot_fig.layout.scene, row=row, col=col)

            n += 1

    return fig



import numpy as np

def tool_create_orthonormal_triad(t):
    """
    This function creates an orthonormal triad given a vector t.
    The function first normalizes the input vector t, then it creates two orthogonal unit vectors e1 and e2.
    The vectors e1, e2, and the normalized t form an orthonormal triad.

    Args:
    t : numpy array
        The input vector. It does not need to be a unit vector.

    Returns:
    e1, e2, t : tuple of numpy arrays
        The orthonormal triad. Each element of the tuple is a numpy array representing a vector.
    """
    t = np.array(t)

    # Normalize the input vector
    t = t / np.linalg.norm(t)

    # Define two orthogonal vectors
    ey = np.array([0, 1, 0])
    ez = np.array([0, 0, 1])

    # Check the dot product of ey and t
    # If it is less than 0.8, we can use ey to create the orthogonal vector e1
    # Otherwise, we use ez to create e1
    if abs(np.dot(ey, t)) < 0.8:
        e1 = np.cross(ey, t)
    else:
        e1 = np.cross(ez, t)

    # Normalize e1
    e1 /= np.linalg.norm(e1)

    # Create the second orthogonal vector e2 by taking the cross product of t and e1
    e2 = np.cross(t, e1)

    # Return the orthonormal triad
    return e1, e2, t



import numpy as np

def tool_complete_field(self, field: np.ndarray) -> np.ndarray:
        """Extends the field in case not a complete array is given. 

        For instance, if a field in 3 dimensions is calculated using only x, then the field is extended to the full 3D array.

        Args:
            field: The field to be extended

        Returns:
            The extended field (np.ndarray)
        """    

        # 2 dimensional fields
        if field.shape == (self.xRes,1):
            field = np.tile(field,(1,self.yRes))

        elif field.shape == (1,self.yRes):
            field = np.tile(field,(self.xRes,1))

        # 3 dimensional fields
        elif field.shape == (self.xRes,1,1):
            field = np.tile(field,(1,self.yRes,1))
            field = np.tile(field,(1,1,self.zRes))

        elif field.shape == (1,self.yRes,1):
            field = np.tile(field,(self.xRes,1,1))
            field = np.tile(field,(1,1,self.zRes))

        elif field.shape == (1,1,self.zRes):
            field = np.tile(field,(self.xRes,1,1))
            field = np.tile(field,(1,self.yRes,1))

        elif field.shape == (self.xRes,self.yRes,1):
            field = np.tile(field,(1,1,self.zRes))

        elif field.shape == (self.xRes,1,self.zRes):
            field = np.tile(field,(1,self.yRes,1))

        elif field.shape == (1,self.yRes,self.zRes):
            field = np.tile(field,(self.xRes,1,1))

        return field



def tool_add_spacing_2D(X,Y,U,V,spacing):
        X = X[::spacing, ::spacing]
        Y = Y[::spacing, ::spacing]
        U = U[::spacing, ::spacing]
        V = V[::spacing, ::spacing]
        return X,Y,U,V


from typing import Union, Optional

import numpy as np
import scipy as sp

from comfit.tool import (
    tool_colormap_angle,
    tool_colormap_bluewhitered,
    tool_colormap_sunburst,
)
from comfit.tool import tool_create_orthonormal_triad
from comfit.tool import tool_set_plot_axis_properties_matplotlib
from comfit.tool import tool_complete_field

from comfit.plot import plot_surface_matplotlib
from comfit.plot import plot_field_matplotlib
from comfit.plot import plot_field_plotly
from comfit.plot import plot_complex_field_matplotlib
from comfit.plot import plot_complex_field_plotly
from comfit.plot import plot_angle_field_matplotlib
from comfit.plot import plot_angle_field_plotly
from comfit.plot import plot_vector_field_matplotlib
from comfit.plot import plot_vector_field_plotly
from comfit.plot import plot_field_in_plane_matplotlib
from comfit.plot import plot_field_in_plane_plotly
from comfit.plot import plot_complex_field_in_plane_matplotlib
from comfit.plot import plot_complex_field_in_plane_plotly
from comfit.plot import plot_angle_field_in_plane_matplotlib
from comfit.plot import plot_angle_field_in_plane_plotly
from comfit.plot import plot_vector_field_in_plane_matplotlib
from comfit.plot import plot_vector_field_in_plane_plotly
from comfit.plot import plot_nodes_matplotlib
from comfit.plot import plot_nodes_plotly

from comfit.plot import plot_subplots_matplotlib
from comfit.plot import plot_subplots_plotly

from comfit.plot import plot_save_matplotlib
from comfit.plot import plot_save_plotly

import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.tri as mtri
import matplotlib.cm as cm
import matplotlib.axes
import matplotlib.figure

import mpl_toolkits

import plotly.graph_objects as go
import plotly.figure_factory as ff
import plotly.colors as pc

from skimage.measure import marching_cubes

import time


class BaseSystemPlot:
    """Plotting methods for the base system class"""

    def plot_field(self, field: np.ndarray, **kwargs):
        if self.plot_lib == "plotly":
            return plot_field_plotly(self, field, **kwargs)
        elif self.plot_lib == "matplotlib":
            return plot_field_matplotlib(self, field, **kwargs)

    def plot_complex_field(self, complex_field: np.ndarray, **kwargs):
        if self.plot_lib == "plotly":
            return plot_complex_field_plotly(self, complex_field, **kwargs)
        elif self.plot_lib == "matplotlib":
            return plot_complex_field_matplotlib(self, complex_field, **kwargs)

    def plot_angle_field(self, angle_field: np.ndarray, **kwargs):
        if self.plot_lib == "plotly":
            return plot_angle_field_plotly(self, angle_field, **kwargs)
        elif self.plot_lib == "matplotlib":
            return plot_angle_field_matplotlib(self, angle_field, **kwargs)

    def plot_vector_field(self, vector_field: np.ndarray, **kwargs):
        if self.plot_lib == "plotly":
            return plot_vector_field_plotly(self, vector_field, **kwargs)
        elif self.plot_lib == "matplotlib":
            return plot_vector_field_matplotlib(self, vector_field, **kwargs)

    def plot_field_in_plane(
        self,
        field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs
    ):
        if self.plot_lib == "plotly":
            return plot_field_in_plane_plotly(
                self, field, normal_vector, position, **kwargs
            )
        elif self.plot_lib == "matplotlib":
            return plot_field_in_plane_matplotlib(
                self, field, normal_vector, position, **kwargs
            )

    def plot_complex_field_in_plane(
        self,
        complex_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs
    ):
        if self.plot_lib == "plotly":
            return plot_complex_field_in_plane_plotly(
                self, complex_field, normal_vector, position, **kwargs
            )
        elif self.plot_lib == "matplotlib":
            return plot_complex_field_in_plane_plotly(
                self, complex_field, normal_vector, position, **kwargs
            )

    def plot_angle_field_in_plane(
        self,
        angle_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs
    ):
        if self.plot_lib == "plotly":
            return plot_angle_field_in_plane_plotly(
                self, angle_field, normal_vector, position, **kwargs
            )
        elif self.plot_lib == "matplotlib":
            return plot_angle_field_in_plane_plotly(
                self, angle_field, normal_vector, position, **kwargs
            )

    def plot_vector_field_in_plane(
        self,
        vector_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        spacing: int = 2,
        **kwargs
    ):
        if self.plot_lib == "plotly":
            return plot_vector_field_in_plane_plotly(
                self, vector_field, normal_vector, position, spacing, **kwargs
            )
        elif self.plot_lib == "matplotlib":
            return plot_vector_field_in_plane_matplotlib(
                self, vector_field, normal_vector, position, spacing, **kwargs
            )

    def plot_nodes(self, disclination_nodes, **kwargs):
        if self.plot_lib == 'plotly':
            return plot_nodes_plotly(self, disclination_nodes, **kwargs)
        elif self.plot_lib == 'matplotlib':
            return plot_nodes_matplotlib(self, disclination_nodes, **kwargs)


    # Figure handling methods
    def plot_subplots(self, *args, **kwargs):
        if self.plot_lib == "plotly":
            return plot_subplots_plotly(number_of_rows, number_of_columns, *args, **kwargs)
        elif self.plot_lib == "matplotlib":
            return plot_subplots_matplotlib(number_of_rows, number_of_columns, *args, **kwargs)

    def plot_save(self, counter, fig, **kwargs):
        if self.plot_lib == "plotly":
            return plot_save_plotly(counter, fig, **kwargs)
        elif self.plot_lib == "matplotlib":
            return plot_save_matplotlib(counter, fig, **kwargs)

    def show(self, fig):
        if self.plot_lib == "matplotlib":
            plt.show()
        elif self.plot_lib == "plotly":
            fig.show()


import numpy as np
import scipy as sp
from typing import Optional
from comfit.tool.tool_complete_field import tool_complete_field
from scipy.interpolate import griddata
import plotly.graph_objects as go
from skimage.measure import marching_cubes
from comfit.tool.tool_set_plot_axis_properties_plotly import tool_set_plot_axis_properties_plotly

def plot_field_in_plane_plotly(
        self,
        field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs
    ):
    
    """Plots the field in a plane perpendicular to the given normal vector
    
    Uses scipy.interpolate.griddata and plt.plot_trisurf.

    Args:
        field (array-like): The field to be plotted.
        normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
        position (array-like, optional): The position of the plane. Default is the middle of the system.
        **kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/. 
    
    Returns:
        The axes containing the plot. (matplotlib.axes.Axes)
    """

    if self.dim != 3:
        raise Exception("The plot in plane function is only defined for 3D fields.")

    kwargs['plot_is_3D'] = True

    # Extend the field if not a complete array is given
    field = tool_complete_field(self, field)

    # Check if the vector field is complex
    if np.iscomplexobj(field):
        print("\033[91mWarning: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(field)))
        field = np.real(field)

    fig = kwargs.get('fig', go.Figure())
    colorbar = kwargs.get('colorbar', True)

    # Default values of position and normal vector
    if position is None:
        position = self.rmid

    if normal_vector is None:
        normal_vector = [0, 1, 0]

    normal_vector = np.array(normal_vector) / np.linalg.norm(normal_vector)
    height_above_plane = (self.x - position[0]) * normal_vector[0] + (self.y - position[1]) * normal_vector[1] + (self.z - position[2]) * normal_vector[2]

    verts, faces, _, _ = marching_cubes(height_above_plane, 0)

    # Calculate the centroids of each triangle
    centroids = np.mean(verts[faces], axis=1)

    # Assuming field is defined on the same grid as height_above_plane
    x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

    # Flatten the grid for interpolation
    points = np.c_[x.ravel(), y.ravel(), z.ravel()]
    field_values = field.ravel()

    # Interpolate field at the vertices positions
    field_verts = griddata(points, field_values, centroids, method='nearest')

    # Add trace
    fig.add_trace(go.Mesh3d(
        x=(self.xmin + verts[:, 0] * self.dx) / self.a0,
        y=(self.ymin + verts[:, 1] * self.dy) / self.a0,
        z=(self.zmin + verts[:, 2] * self.dz) / self.a0,
        i=faces[:, 0],
        j=faces[:, 1],
        k=faces[:, 2],
        intensity=field_verts,  
        intensitymode='cell',  
        colorscale='Viridis',
        showscale=colorbar
    ))

    kwargs['fig'] = fig
    tool_set_plot_axis_properties_plotly(self, **kwargs)
    return fig
